import abc
import typing
import wtforms.validators
from starlette.authentication import AuthCredentials, AuthenticationBackend, BaseUser, UnauthenticatedUser
from starlette.requests import HTTPConnection, Request
from starlette_babel import gettext_lazy as _

SESSION_KEY = '_auth_user_id_'


class AdminUser(typing.Protocol):
    display_name: str
    avatar: str


class LoginForm(wtforms.Form):
    identity = wtforms.EmailField(
        label=_('Email', domain='ohmyadmin'),
        render_kw={'autocomplete': 'email', 'inputmode': 'email'},
        validators=[
            wtforms.validators.data_required(),
        ],
    )
    password = wtforms.PasswordField(
        label=_('Password', domain='ohmyadmin'),
        render_kw={'autocomplete': 'password'},
        validators=[
            wtforms.validators.data_required(),
        ],
    )
    next_url = wtforms.HiddenField()


class AnonymousUser(UnauthenticatedUser):
    is_anonymous = True
    display_name = _('Anonymous', domain='ohmyadmin')


class BaseAuthPolicy(abc.ABC):
    login_form_class: type[LoginForm] = LoginForm

    @abc.abstractmethod
    async def authenticate(self, request: Request, identity: str, password: str) -> BaseUser | None:  # pragma: nocover
        ...

    @abc.abstractmethod
    async def load_user(self, conn: HTTPConnection, user_id: str) -> BaseUser | None:  # pragma: nocover
        ...

    def login(self, request: Request, user: BaseUser) -> None:
        request.session[SESSION_KEY] = user.identity

    def logout(self, request: Request) -> None:
        if SESSION_KEY in request.session:
            del request.session[SESSION_KEY]

    def get_login_form_class(self) -> type[LoginForm]:
        return self.login_form_class

    def get_authentication_backend(self) -> AuthenticationBackend:
        return SessionAuthBackend()


class AnonymousAuthPolicy(BaseAuthPolicy):
    async def authenticate(self, request: Request, identity: str, password: str) -> BaseUser | None:
        return None

    async def load_user(self, conn: HTTPConnection, user_id: str) -> BaseUser | None:
        return AnonymousUser()


class SessionAuthBackend(AuthenticationBackend):
    async def authenticate(self, conn: HTTPConnection) -> tuple[AuthCredentials, BaseUser] | None:
        auth_policy: BaseAuthPolicy = conn.state.admin.auth_policy
        user_id = conn.session.get(SESSION_KEY, '')
        if user_id and (user := await auth_policy.load_user(conn, user_id)):
            return AuthCredentials(), user
        return AuthCredentials([]), UnauthenticatedUser()



import functools
import typing
from markupsafe import Markup
from starlette.requests import Request
from unittest import mock

from ohmyadmin import actions
from ohmyadmin.formatters import DataFormatter, StringFormatter
from ohmyadmin.helpers import LazyObjectURL, LazyURL, snake_to_sentence
from ohmyadmin.ordering import SortingHelper
from ohmyadmin.pagination import Pagination
from ohmyadmin.shortcuts import render_to_string
from ohmyadmin.views.base import IndexView


def default_value_getter(obj: typing.Any, attr: str) -> typing.Any:
    return getattr(obj, attr, 'undefined!')


class TableColumn:
    def __init__(
        self,
        name: str,
        label: str | None = None,
        searchable: bool = False,
        sortable: bool = False,
        formatter: DataFormatter | None = None,
        search_in: str | None = None,
        sort_by: str | None = None,
        link: bool | str | LazyURL | LazyObjectURL | None = None,
        value_getter: typing.Callable[[typing.Any], typing.Any] | None = None,
    ) -> None:
        self.link = link
        self.name = name
        self.sortable = sortable
        self.searchable = searchable
        self.search_in = search_in or name
        self.sort_by = sort_by or name
        self.formatter = formatter or StringFormatter()
        self.label = label or snake_to_sentence(name)
        self.value_getter = value_getter or functools.partial(default_value_getter, attr=self.name)

    def get_value(self, obj: typing.Any) -> typing.Any:
        return self.value_getter(obj)

    def format_value(self, request: Request, value: typing.Any) -> str:
        return self.formatter(request, value)

    def render(self, request: Request, obj: typing.Any) -> str:
        value = self.get_value(obj)
        display_value = self.format_value(request, value)
        link: str = ''
        match self.link:
            case str():
                link = self.link
            case bool():
                link = request.state.page.generate_url(request)
            case LazyURL():
                link = str(self.link.resolve(request))
            case LazyObjectURL():
                link = str(self.link.resolve(request, obj))

        return request.state.admin.render_to_string(
            request,
            'ohmyadmin/views/table/table_cell.html',
            {
                'link': link,
                'column': self,
                'value': display_value,
            },
        )

    def render_head_cell(self, request: Request) -> str:
        sorting: SortingHelper = request.state.table_sorting
        return request.state.admin.render_to_string(
            request,
            'ohmyadmin/views/table/table_head_cell.html',
            {
                'column': self,
                'sorting': sorting,
            },
        )


class TableView(IndexView):
    def __init__(self, columns: typing.Sequence[TableColumn]) -> None:
        self.columns = columns

    def get_object_actions(self, request: Request, obj: typing.Any) -> typing.Sequence[actions.ObjectAction]:
        return request.state.page.get_object_actions(request, obj)

    def render(self, request: Request, objects: Pagination[typing.Any]) -> str:
        request.state.table_sorting = SortingHelper(request, request.state.page.ordering_param)
        has_object_actions = self.get_object_actions(request, mock.MagicMock())
        has_batch_actions = bool(request.state.page.get_batch_actions(request))
        return Markup(
            render_to_string(
                request,
                'ohmyadmin/views/table/table.html',
                {
                    'table': self,
                    'objects': objects,
                    'has_batch_actions': has_batch_actions,
                    'has_object_actions': has_object_actions,
                },
            )
        )



import sqlalchemy as sa
import typing
from starlette.requests import Request

from ohmyadmin.forms import AsyncChoicesLoader


def choices_from(
    model_class: typing.Any,
    value_column: str = 'id',
    label_column: str = 'name',
    query: sa.Select | None = None,
) -> AsyncChoicesLoader:
    query = query or sa.select(model_class)

    async def loader(request: Request) -> typing.Sequence[tuple[typing.Any, str]]:
        result = await request.state.dbsession.scalars(query)
        choices: list[tuple[typing.Any, str]] = []
        for row in result:
            choices.append((getattr(row, value_column), getattr(row, label_column)))
        return choices

    return loader



import bs4
import typing
from bs4 import BeautifulSoup


class SelectorError(Exception):
    ...


class NodeNotFoundError(SelectorError):
    ...


class MarkupSelector:
    def __init__(self, markup: str | bytes) -> None:
        if isinstance(markup, bytes):
            markup = markup.decode()
        self.root = BeautifulSoup(markup, multi_valued_attributes=False, features="html.parser")

    def find_node(self, selector: str) -> bs4.Tag | None:
        return self.root.select_one(selector)

    def find_node_or_raise(self, selector: str) -> bs4.Tag:
        if node := self.root.select_one(selector):
            return node
        raise NodeNotFoundError(f'No nodes: "{selector}".')

    def get_attribute(self, selector: str, attribute: str, default: typing.Any = None) -> str:
        return self.find_node_or_raise(selector).get(attribute, default)

    def has_attribute(self, selector: str, attribute: str) -> bool:
        return attribute in self.find_node_or_raise(selector).attrs

    def get_classes(self, selector: str) -> list[str]:
        return self.find_node_or_raise(selector).attrs.get('class', '').split(' ')

    def has_class(self, selector: str, *class_name: str) -> bool:
        classes = set(self.get_classes(selector))
        return len(classes.intersection(set(class_name))) == len(class_name)

    def get_styles(self, selector: str) -> dict[str, str]:
        return {
            css[0]: css[1]
            for css in [
                list(map(str.strip, style.split(':')))
                for style in self.get_attribute(selector, 'style', '').split(';')
                if style
            ]
        }

    def get_style(self, selector: str, name: str) -> str:
        return self.get_styles(selector).get(name, '')

    def has_style(self, selector: str, name: str) -> bool:
        return name in self.get_styles(selector)

    def get_text(self, selector: str) -> str:
        node = self.find_node(selector)
        if node is None:
            return ''
        return node.text.strip()

    def has_node(self, selector: str) -> bool:
        return self.find_node(selector) is not None

    def count(self, selector: str) -> int:
        return len(self.root.select(selector))



import abc
import typing
from starlette.requests import Request

from ohmyadmin.pagination import Pagination


class IndexView:
    @abc.abstractmethod
    def render(self, request: Request, objects: Pagination[typing.Any]) -> str:  # pragma: no cover
        ...



import abc
import inspect
import typing
import wtforms
from starlette.concurrency import run_in_threadpool
from starlette.requests import Request

_F = typing.TypeVar('_F', bound=wtforms.Form)


async def create_form(request: Request, form_class: type[_F], obj: typing.Any | None = None) -> _F:
    form_data = None if request.method in ['GET', 'HEAD', 'OPTIONS'] else await request.form()
    form = form_class(form_data, obj=obj)
    await init_form(request, form)
    return form


def iterate_form_fields(form: typing.Iterable[wtforms.Field]) -> typing.Generator[wtforms.Field, None, None]:
    for field in form:
        if isinstance(field, (wtforms.FieldList, wtforms.FormField)):
            yield from iterate_form_fields(field)
        else:
            yield field


async def init_form(request: Request, form: wtforms.Form) -> None:
    for field in iterate_form_fields(form):
        if isinstance(field, Initable):
            await field.init(request)


async def validate_form(form: wtforms.Form) -> bool:
    """
    Perform form validation.

    This function does not call Form.validate or Field.validate, instead it implements own logic that supports async
    validators.
    """
    is_valid = True
    for field in iterate_form_fields(form):
        for validator in field.validators:
            field.errors = list(field.errors) if field.errors is not None else []
            try:
                if inspect.iscoroutinefunction(validator):
                    await validator(form, field)
                else:
                    await run_in_threadpool(validator, form, field)
            except (wtforms.ValidationError, wtforms.validators.StopValidation) as ex:
                field.errors = list(field.errors)
                field.errors.extend(ex.args)
                is_valid = False
    return is_valid


async def validate_on_submit(request: Request, form: wtforms.Form) -> bool:
    if request.method in ['POST', 'PUT', 'PATCH', 'DELETE']:
        return await validate_form(form)
    return False


async def populate_object(request: Request, form: wtforms.Form, obj: typing.Any) -> None:
    form.populate_obj(obj)


class Initable(abc.ABC):  # pragma: no cover
    @abc.abstractmethod
    async def init(self, request: Request) -> None:
        ...


class Processable(abc.ABC):  # pragma: no cover
    @abc.abstractmethod
    async def process(self, request: Request) -> None:
        ...


Choices = typing.Sequence[tuple[typing.Any, str]]


class AsyncChoicesLoader(typing.Protocol):  # pragma: no cover
    async def __call__(self, request: Request) -> Choices:
        ...


ChoicesType = typing.Union[
    Choices,
    AsyncChoicesLoader,
    typing.Callable[[], Choices],
]


class AsyncSelectField(wtforms.SelectField, Initable):
    def __init__(self, choices: ChoicesType, **kwargs: typing.Any) -> None:
        self.choices_loader = None
        if inspect.iscoroutinefunction(choices):
            self.choices_loader = choices
            choices = []
        super().__init__(choices=choices, **kwargs)

    async def init(self, request: Request) -> None:
        if self.choices_loader:
            self.choices = list(await self.choices_loader(request))


class AsyncSelectMultipleField(wtforms.SelectMultipleField, Initable):
    def __init__(self, choices: ChoicesType, **kwargs: typing.Any) -> None:
        self.choices_loader = None
        if inspect.iscoroutinefunction(choices):
            self.choices_loader = choices
            choices = []
        super().__init__(choices=choices, **kwargs)

    async def init(self, request: Request) -> None:
        if self.choices_loader:
            self.choices = list(await self.choices_loader(request))


_SCPS = typing.ParamSpec('_SCPS')
_SCRT = typing.TypeVar('_SCRT')


def safe_coerce(callback: typing.Callable[[typing.Any], _SCRT]) -> typing.Callable[[typing.Any], _SCRT | None]:
    def coerce(value: float | typing.AnyStr) -> _SCRT | None:
        if value is None:
            return None

        try:
            return callback(value)
        except ValueError:
            return None

    return coerce



import datetime
import decimal
import typing
from starlette.requests import Request
from starlette_babel import gettext_lazy as _

from ohmyadmin.shortcuts import render_to_string

TextAlign = typing.Literal['left', 'right', 'center']
_T = typing.TypeVar('_T')


class DataFormatter(typing.Protocol):  # pragma: no cover
    """
    Data formatter is a utility that renders object field value into a different way.

    For example, boolean can be rendered as icons or yes/no text.
    """

    def __call__(self, request: Request, value: typing.Any) -> str:
        ...


class BaseFormatter(typing.Generic[_T]):
    template: str = 'ohmyadmin/formatters/string.html'

    def format(self, request: Request, value: _T) -> str:
        return render_to_string(request, self.template, {'formatter': self, 'value': value})

    def __call__(self, request: Request, value: _T) -> str:
        return self.format(request, value)


class StringFormatter(BaseFormatter[str]):
    pass


class DateTimeFormatter(BaseFormatter[datetime.datetime]):
    template: str = 'ohmyadmin/formatters/datetime.html'

    def __init__(self, format: typing.Literal['short', 'medium', 'long', 'full'] = 'medium') -> None:
        self.value_format = format


class DateFormatter(BaseFormatter[datetime.date | datetime.datetime]):
    template: str = 'ohmyadmin/formatters/date.html'

    def __init__(self, format: typing.Literal['short', 'medium', 'long', 'full'] = 'medium') -> None:
        self.value_format = format


class TimeFormatter(BaseFormatter[datetime.datetime | datetime.time]):
    template: str = 'ohmyadmin/formatters/time.html'

    def __init__(self, format: typing.Literal['short', 'medium', 'long', 'full'] = 'short') -> None:
        self.value_format = format


class BoolFormatter(BaseFormatter[bool]):
    template: str = 'ohmyadmin/formatters/bool.html'

    def __init__(
        self,
        as_text: bool = False,
        true_text: str = _('Yes', domain='ohmyadmin'),
        false_text: str = _('No', domain='ohmyadmin'),
    ) -> None:
        self.true_text = true_text
        self.false_text = false_text
        self.as_text = as_text


class AvatarFormatter(BaseFormatter[str]):
    template: str = 'ohmyadmin/formatters/avatar.html'


class NumberFormatter(BaseFormatter[int | float | decimal.Decimal]):
    template: str = 'ohmyadmin/formatters/number.html'

    def __init__(self, *, prefix: str = '', suffix: str = '', align: TextAlign = 'right') -> None:
        self.prefix = prefix
        self.suffix = suffix
        self.align = align


BadgeColor: typing.TypeAlias = typing.Literal['gray', 'red', 'yellow', 'green', 'blue', 'indigo', 'purple', 'pink']


class BadgeFormatter(BaseFormatter[str | int]):
    template: str = 'ohmyadmin/formatters/badge.html'

    def __init__(self, *, color_map: dict[str | int, BadgeColor]) -> None:
        self.color_map = color_map or {}

    def format(self, request: Request, value: str | int) -> str:
        color = self.color_map.get(value, 'gray')
        return render_to_string(request, self.template, {'formatter': self, 'value': value, 'color': color})


ProgressSize = typing.Literal['xxs', 'xs', 'sm', 'md', 'lg']
ProgressColor = typing.Literal['accent', 'gray', 'red', 'yellow', 'green', 'blue', 'indigo', 'purple', 'pink']


class ProgressFormatter(BaseFormatter[int | float]):
    template: str = 'ohmyadmin/formatters/progress.html'

    def __init__(
        self,
        *,
        size: ProgressSize = 'sm',
        color: ProgressColor = 'accent',
        label: str = '',
    ) -> None:
        self.size = size
        self.color = color
        self.label = label



import dataclasses

import inspect
import re
import typing
from starlette.datastructures import URL
from starlette.requests import Request


def camel_to_sentence(text: str) -> str:
    """
    Convert camel cased strings (like class names) into a sentence.

    Example: OrderItemsResource -> Order items resource.
    """
    return re.sub(r'(?<!^)(?=[A-Z])', ' ', text).lower().capitalize()


def snake_to_sentence(text: str) -> str:
    """Convert snake_case string into 'Snake case' sentence."""
    return text.replace('_', ' ').lower().capitalize()


uncountable = [
    'audio',
    'bison',
    'cattle',
    'chassis',
    'compensation',
    'coreopsis',
    'data',
    'deer',
    'education',
    'emoji',
    'equipment',
    'evidence',
    'feedback',
    'firmware',
    'fish',
    'furniture',
    'gold',
    'hardware',
    'information',
    'jedi',
    'kin',
    'knowledge',
    'love',
    'metadata',
    'money',
    'moose',
    'news',
    'nutrition',
    'offspring',
    'plankton',
    'pokemon',
    'police',
    'rain',
    'rice',
    'series',
    'sheep',
    'software',
    'species',
    'swine',
    'traffic',
    'wheat',
]


def pluralize(text: str) -> str:
    """
    A simple pluralization utility.

    It adds -ies suffix to any noun that ends with -y, it adds -es suffix to any noun that ends with -s. For all other
    cases it appends -s.
    """
    if text in uncountable:
        return text

    # is already plural?
    if text.endswith('ies') or text.endswith('rs') or text.endswith('ds'):
        return text

    if text.endswith('y'):
        return text[:-1] + 'ies'
    if text.endswith('s'):
        return text + 'es'
    return text + 's'


@dataclasses.dataclass
class LazyURL:
    path_name: str
    path_params: dict[str, typing.Any] = dataclasses.field(default_factory=dict)

    def resolve(self, request: Request) -> URL:
        return URL(request.url_for(self.path_name, **self.path_params))


class LazyObjectURL:
    def __init__(self, factory: typing.Callable[[Request, typing.Any], URL]) -> None:
        self.factory = factory

    def resolve(self, request: Request, obj: typing.Any) -> URL:
        return self.factory(request, obj)


def resolve_url(request: Request, url: str | URL | LazyURL) -> URL:
    if isinstance(url, LazyURL):
        return url.resolve(request)
    return URL(str(url))


def get_callable_name(obj: typing.Any) -> str:
    if inspect.isfunction(obj):
        return obj.__name__
    if inspect.isclass(obj):
        return obj.__name__
    if inspect.ismethod(obj):
        return obj.__name__
    raise ValueError('Unsupported')



import abc
import typing
import wtforms
from starlette.requests import Request

from ohmyadmin.shortcuts import render_to_string


class Layout(abc.ABC):
    @abc.abstractmethod
    def render(self, request: Request) -> str:  # pragma: no cover
        ...


class TemplateLayout(Layout):
    template: str

    def render(self, request: Request) -> str:
        assert self.template, 'template is undefined'
        return render_to_string(request, self.template, {'layout': self})


class Grid(TemplateLayout):
    template = 'ohmyadmin/layouts/grid.html'

    def __init__(self, children: typing.Sequence[Layout], columns: int = 2, gap: int = 6) -> None:
        self.gap = gap
        self.columns = columns
        self.children = children


class Column(TemplateLayout):
    template = 'ohmyadmin/layouts/column.html'

    def __init__(self, children: typing.Sequence[Layout], gap: int = 5, colspan: int | str = '') -> None:
        self.gap = gap
        self.colspan = colspan
        self.children = children


class Row(TemplateLayout):
    template = 'ohmyadmin/layouts/row.html'

    def __init__(self, children: typing.Sequence[Layout], gap: int = 5) -> None:
        self.gap = gap
        self.children = children


class Card(TemplateLayout):
    template = 'ohmyadmin/layouts/card.html'

    def __init__(self, children: typing.Sequence[Layout], label: str = '', description: str = '') -> None:
        self.label = label
        self.children = children
        self.description = description


class SideSection(TemplateLayout):
    template = 'ohmyadmin/layouts/fieldset.html'

    def __init__(self, children: typing.Sequence[Layout], label: str = '', description: str = '') -> None:
        self.label = label
        self.children = children
        self.description = description


class Input(TemplateLayout):
    template = 'ohmyadmin/layouts/input.html'

    def __init__(self, field: wtforms.Field, max_width: str = 'full', colspan: str | int = '') -> None:
        self.field = field
        self.colspan = colspan
        self.max_width = max_width


class RowFormLayout(TemplateLayout):
    template = 'ohmyadmin/layouts/simple_form.html'

    def __init__(self, children: typing.Sequence[Input], label: str = '', description: str = '') -> None:
        self.children = children
        self.label = label
        self.description = description


class StackedForm(TemplateLayout):
    template = 'ohmyadmin/layouts/stacked_form.html'

    def __init__(self, children: typing.Sequence[Input], label: str = '', description: str = '') -> None:
        self.children = children
        self.label = label
        self.description = description


class Text(Layout):
    def __init__(self, content: str) -> None:
        self.content = content

    def render(self, request: Request) -> str:
        return self.content



from ohmyadmin.contrib.sqlalchemy.datasource import SQLADataSource
from ohmyadmin.contrib.sqlalchemy.middleware import DatabaseSessionMiddleware

__all__ = ['DatabaseSessionMiddleware', 'SQLADataSource']



from sqlalchemy.ext.asyncio import async_sessionmaker
from starlette.types import ASGIApp, Receive, Scope, Send


class DatabaseSessionMiddleware:
    def __init__(self, app: ASGIApp, async_session: async_sessionmaker) -> None:
        self.app = app
        self.async_session = async_session

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        async with self.async_session() as session:
            scope.setdefault('state', {})
            scope['state']['dbsession'] = session
            await self.app(scope, receive, send)



import typing
import wtforms
from starlette.datastructures import URL, FormData
from starlette.requests import Request
from starlette.responses import RedirectResponse, Response
from starlette.routing import BaseRoute, Mount, Route, Router
from starlette.types import Receive, Scope, Send
from starlette_babel import gettext_lazy as _
from starlette_flash import flash

from ohmyadmin import actions, layouts
from ohmyadmin.actions import ActionResponse, BatchDelete, DeleteObjectAction
from ohmyadmin.datasource.base import DataSource
from ohmyadmin.filters import BaseFilter, UnboundFilter
from ohmyadmin.forms import populate_object
from ohmyadmin.helpers import LazyObjectURL, LazyURL
from ohmyadmin.metrics import Card
from ohmyadmin.pages.base import BasePage
from ohmyadmin.pages.form import FormPage
from ohmyadmin.pages.table import TablePage
from ohmyadmin.views.table import TableColumn


class Resource(BasePage, Router):
    group = _('Resources', domain='ohmyadmin')
    form_class: type[wtforms.Form] = wtforms.Form
    datasource: DataSource
    page_param: typing.ClassVar[str] = 'page'
    page_size_param: typing.ClassVar[str] = 'page_size'
    search_param: typing.ClassVar[str] = 'search'
    ordering_param: typing.ClassVar[str] = 'ordering'
    page_size: typing.ClassVar[int] = 25
    max_page_size: typing.ClassVar[int] = 100
    columns: typing.Sequence[TableColumn] | None = None
    page_actions: typing.Sequence[actions.PageAction] | None = None
    object_actions: typing.Sequence[actions.ObjectAction] | None = None
    batch_actions: typing.Sequence[actions.BatchAction] | None = None
    filters: typing.Sequence[UnboundFilter | BaseFilter] | None = None
    metrics: typing.Sequence[type[Card]] | None = None
    create_form_actions: typing.Sequence[actions.Submit | actions.Link] | None = None
    update_form_actions: typing.Sequence[actions.Submit | actions.Link] | None = None

    def __init__(self) -> None:
        super().__init__(routes=self.get_routes())

    def get_table_columns(self, request: Request) -> typing.Sequence[TableColumn]:
        columns = list(self.columns or [])
        for column in columns:
            if column.link is True:
                column.link = LazyObjectURL(
                    lambda r, o: URL(request.url_for(self.get_path_name() + '.edit', pk=self.datasource.get_pk(o)))
                )

        return columns

    def get_page_actions(self, request: Request) -> list[actions.PageAction]:
        create_route_name = self.get_path_name() + '.create'
        return [
            *(self.page_actions or []),
            actions.Link(
                label=_('Create', domain='ohmyadmin'),
                url=LazyURL(create_route_name),
                variant='accent',
                icon='plus',
            ),
        ]

    def get_object_actions(self, request: Request, obj: typing.Any) -> list[actions.ObjectAction]:
        edit_route_name = self.get_path_name() + '.edit'
        return [
            *(self.object_actions or []),
            actions.ObjectLink(
                label=_('Edit', domain='ohmyadmin'),
                icon='pencil',
                url=LazyURL(path_name=edit_route_name, path_params={'pk': self.datasource.get_pk(obj)}),
            ),
            DeleteObjectAction(),
        ]

    def get_batch_actions(self, request: Request) -> list[actions.BatchAction]:
        return [*(self.batch_actions or []), BatchDelete()]

    def get_filters(self, request: Request) -> typing.Sequence[UnboundFilter | BaseFilter]:
        return self.filters or []

    def get_metrics(self, request: Request) -> typing.Sequence[Card]:
        return [metric() for metric in self.metrics or []]

    def build_form_layout(self, request: Request, form: wtforms.Form) -> layouts.Layout:
        return layouts.Card([layouts.StackedForm([layouts.Input(field) for field in form])])

    def get_create_form_actions(self, request: Request) -> typing.Sequence[actions.Submit | actions.Link]:
        return self.create_form_actions or [
            actions.Submit(label=_('Create and return to list', domain='ohmyadmin'), variant='accent', name='_return'),
            actions.Submit(label=_('Create and edit', domain='ohmyadmin'), name='_edit'),
            actions.Submit(label=_('Create and add new', domain='ohmyadmin'), name='_add_new'),
        ]

    def get_create_view_response(self, request: Request, form_data: FormData, model: typing.Any) -> Response:
        flash(request).success(_('{object} has been created.', domain='ohmyadmin').format(object=model))
        if '_edit' in form_data:
            return ActionResponse().redirect(
                request,
                request.url_for(
                    name=self.get_path_name() + '.edit',
                    pk=self.datasource.get_pk(model),
                ),
            )
        if '_add_new' in form_data:
            return ActionResponse().redirect(request, request.url_for(name=self.get_path_name() + '.create'))
        return ActionResponse().redirect(request, self.generate_url(request))

    def get_update_form_actions(self, request: Request) -> typing.Sequence[actions.Submit | actions.Link]:
        return self.update_form_actions or [
            actions.Submit(label=_('Update and return to list', domain='ohmyadmin'), variant='accent', name='_return'),
            actions.Submit(label=_('Update and continue editing', domain='ohmyadmin'), name='_continue'),
            actions.Link(label=_('Return to list', domain='ohmyadmin'), url=self.generate_url(request)),
        ]

    def get_update_view_response(self, request: Request, form_data: FormData, model: typing.Any) -> Response:
        message = _('{object} has been updated.', domain='ohmyadmin').format(object=model)
        if '_continue' not in form_data:
            flash(request).success(_('{object} has been updated.', domain='ohmyadmin').format(object=model))
            return ActionResponse().redirect(request, self.generate_url(request))
        return ActionResponse().show_toast(message)

    async def index_view(self, request: Request) -> Response:
        page_class = type(
            'IndexPage',
            (TablePage,),
            dict(
                page_title=self.page_title,
                label=self.label,
                label_plural=self.label_plural,
                datasource=self.datasource,
                page_param=self.page_param,
                page_size=self.page_size,
                page_size_param=self.page_size_param,
                search_param=self.search_param,
                ordering_param=self.ordering_param,
                max_page_size=self.max_page_size,
                filters=self.get_filters(request),
                get_metrics=self.get_metrics,
                get_page_actions=self.get_page_actions,
                get_object_actions=self.get_object_actions,
                get_batch_actions=self.get_batch_actions,
                get_table_columns=self.get_table_columns,
            ),
        )

        page = page_class()
        return await page.dispatch(request)

    async def create_empty_model(self, request: Request) -> typing.Any:
        return self.datasource.new()

    async def get_object(self, request: Request) -> typing.Any:
        pk = request.path_params['pk']
        return await self.datasource.get(request, pk)

    async def perform_create(self, request: Request, form: wtforms.Form, model: typing.Any) -> Response:
        await populate_object(request, form, model)
        await self.datasource.create(request, model)
        form_data = await request.form()
        return self.get_create_view_response(request, form_data, model)

    async def perform_update(self, request: Request, form: wtforms.Form, model: typing.Any) -> Response:
        await populate_object(request, form, model)
        await self.datasource.update(request, model)
        form_data = await request.form()
        return self.get_update_view_response(request, form_data, model)

    async def create_view(self, request: Request) -> Response:
        page_class = type(
            'CreateFormPage',
            (FormPage,),
            dict(
                label='Create {label_singular}'.format(label_singular=self.label),
                form_class=self.form_class,
                form_actions=self.get_create_form_actions(request),
                build_form_layout=self.build_form_layout,
                get_form_object=self.create_empty_model,
                handle_submit=self.perform_create,
            ),
        )

        page = page_class()
        return await page.dispatch(request)

    async def update_view(self, request: Request) -> Response:
        page_class = type(
            'UpdateFormPage',
            (FormPage,),
            dict(
                label='Update {label_singular}'.format(label_singular=self.label),
                form_class=self.form_class,
                form_actions=self.get_update_form_actions(request),
                build_form_layout=self.build_form_layout,
                get_form_object=self.get_object,
                handle_submit=self.perform_update,
            ),
        )

        page = page_class()
        return await page.dispatch(request)

    async def delete_view(self, request: Request) -> Response:
        pk = request.path_params['pk']
        model = await self.datasource.get(request, pk)
        if request.method in ['POST', 'DELETE']:
            await self.datasource.delete(request, pk)
            flash(request).success(_('{object} has been deleted.', domain='ohmyadmin'))

            redirect_to = self.generate_url(request)
            if 'hx-request' in request.headers:
                return ActionResponse().redirect(request, redirect_to)
            return RedirectResponse(redirect_to, 302)

        cancel_url = self.generate_url(request)
        return self.render_to_response(
            request,
            'ohmyadmin/resources/delete.html',
            {
                'page_title': _('Delete {object}', domain='ohmyadmin').format(object=model),
                'object': model,
                'cancel_url': cancel_url,
            },
        )

    def get_routes(self) -> list[BaseRoute]:
        return [
            Route(
                '/',
                self.index_view,
                name=f'{self.get_path_name()}.index',
                methods=['get', 'post', 'put', 'patch', 'delete'],
            ),
            Route('/new', self.create_view, name=f'{self.get_path_name()}.create', methods=['get', 'post']),
            Route('/{pk}/edit', self.update_view, name=f'{self.get_path_name()}.edit', methods=['get', 'post']),
            Route(
                '/{pk}/delete',
                self.delete_view,
                name=f'{self.get_path_name()}.delete',
                methods=['get', 'post', 'delete'],
            ),
        ]

    def as_route(self) -> Mount:
        return Mount(f'/resources/{self.slug}', self)

    @classmethod
    def generate_url(cls, request: Request) -> str:
        return request.url_for(f'{cls.get_path_name()}.index')

    @classmethod
    def get_path_name(cls) -> str:
        return f'ohmyadmin.resources.{cls.slug}'

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        scope['state']['page'] = self
        scope['state']['datasource'] = self.datasource
        return await super().__call__(scope, receive, send)



from __future__ import annotations

import abc
import datetime
import decimal
import enum
import typing
from starlette.requests import Request
from starlette_babel import gettext_lazy as _

from ohmyadmin.ordering import SortingType
from ohmyadmin.pagination import Pagination


class DataSourceError(Exception):
    """Common exception class for data source related errors."""


class StringOperation(enum.Enum):
    exact = _('same as', domain='ohmyadmin')
    startswith = _('starts with', domain='ohmyadmin')
    endswith = _('ends with', domain='ohmyadmin')
    contains = _('contains', domain='ohmyadmin')
    pattern = _('matches', domain='ohmyadmin')

    @classmethod
    def choices(cls) -> typing.Sequence[tuple[str, str]]:
        return [(choice.name, choice.value) for choice in cls]


class NumberOperation(enum.Enum):
    eq = _('equals', domain='ohmyadmin')
    gt = _('is greater than', domain='ohmyadmin')
    gte = _('is greater than or equal', domain='ohmyadmin')
    lt = _('is less than', domain='ohmyadmin')
    lte = _('is less than or equal', domain='ohmyadmin')

    @classmethod
    def choices(cls) -> typing.Sequence[tuple[str, str]]:
        return [(choice.name, choice.value) for choice in cls]


T = typing.TypeVar('T')


class DataSource(abc.ABC, typing.Generic[T]):
    @abc.abstractmethod
    def get_query_for_index(self) -> DataSource:  # pragma: no cover
        ...

    @abc.abstractmethod
    def get_pk(self, obj: typing.Any) -> str:  # pragma: no cover
        ...

    @abc.abstractmethod
    def new(self) -> typing.Any:  # pragma: no cover
        ...

    @abc.abstractmethod
    def apply_search(
        self, search_term: str, searchable_fields: typing.Sequence[str]
    ) -> DataSource[T]:  # pragma: no cover
        ...

    @abc.abstractmethod
    def apply_ordering(
        self, ordering: dict[str, SortingType], sortable_fields: typing.Sequence[str]
    ) -> DataSource[T]:  # pragma: no cover
        ...

    @abc.abstractmethod
    def apply_string_filter(
        self, field: str, operation: StringOperation, value: str
    ) -> DataSource[T]:  # pragma: no cover
        ...

    @abc.abstractmethod
    def apply_number_filter(
        self, field: str, operation: NumberOperation, value: int | float | decimal.Decimal
    ) -> DataSource[T]:  # pragma: no cover
        ...

    @abc.abstractmethod
    def apply_date_filter(self, field: str, value: datetime.date) -> DataSource[T]:  # pragma: no cover
        ...

    @abc.abstractmethod
    def apply_date_range_filter(
        self, field: str, before: datetime.date | None, after: datetime.date | None
    ) -> DataSource[T]:  # pragma: no cover
        ...

    @abc.abstractmethod
    def apply_choice_filter(
        self, field: str, choices: list[typing.Any], coerce: typing.Callable
    ) -> DataSource[T]:  # pragma: no cover
        ...

    @abc.abstractmethod
    def apply_boolean_filter(self, field: str, value: bool) -> DataSource[T]:  # pragma: no cover
        pass

    @abc.abstractmethod
    async def get(self, request: Request, pk: str) -> T | None:  # pragma: no cover
        ...

    @abc.abstractmethod
    async def paginate(self, request: Request, page: int, page_size: int) -> Pagination[T]:  # pragma: no cover
        ...

    @abc.abstractmethod
    async def create(self, request: Request, model: typing.Any) -> None:  # pragma: no cover
        ...

    @abc.abstractmethod
    async def update(self, request: Request, model: typing.Any) -> None:  # pragma: no cover
        ...

    @abc.abstractmethod
    async def delete(self, request: Request, *object_ids: str) -> None:  # pragma: no cover
        ...



from __future__ import annotations

import abc
import json
import typing
import wtforms
from slugify import slugify
from starlette.background import BackgroundTask
from starlette.datastructures import URL, MultiDict
from starlette.requests import Request
from starlette.responses import Response
from starlette_babel import gettext_lazy as _
from urllib.parse import parse_qsl, urlencode

from ohmyadmin.datasource.base import DataSource
from ohmyadmin.forms import create_form, validate_on_submit
from ohmyadmin.helpers import LazyURL, resolve_url
from ohmyadmin.shortcuts import render_to_response, render_to_string

ButtonType = typing.Literal['default', 'accent', 'primary', 'danger', 'warning', 'success', 'text', 'link']
RequestMethod = typing.Literal['get', 'post', 'patch', 'put', 'delete']
ObjectURLFactory = typing.Callable[[Request, typing.Any], URL | str]


class ActionError(Exception):
    ...


class UndefinedActionError(ActionError):
    ...


class ActionResponse(Response):
    def __init__(
        self,
        status_code: int = 204,
        headers: dict[str, typing.Any] | None = None,
        background: BackgroundTask | None = None,
    ) -> None:
        super().__init__(status_code=status_code, headers=headers, background=background)

    def show_toast(self, message: str, category: typing.Literal['success', 'error'] = 'success') -> ActionResponse:
        return self.trigger('toast', {'message': str(message), 'category': category})

    def redirect(self, request: Request, url: str | LazyURL) -> ActionResponse:
        """Triggers a client-side redirect to a new location."""
        self.headers['hx-redirect'] = str(url.resolve(request)) if isinstance(url, LazyURL) else url
        return self

    def refresh(self) -> ActionResponse:
        """Trigger full refresh of the page."""
        self.headers['hx-refresh'] = 'true'
        return self

    def refresh_datatable(self) -> ActionResponse:
        return self.trigger('refresh-datatable')

    def close_modal(self) -> ActionResponse:
        return self.trigger('modals.close')

    def trigger(self, event: str, data: str | dict[str, str | float] = '') -> ActionResponse:
        payload = json.loads(self.headers.get('hx-trigger', '{}'))
        payload[event] = data
        self.headers['hx-trigger'] = json.dumps(payload)
        return self


response = ActionResponse

ActionCallback = typing.Callable[[Request], typing.Awaitable[Response]]


class PageAction(abc.ABC):
    template: str = ''

    def render(self, request: Request) -> str:
        assert self.template, f'template is not defined on class {self.__class__.__name__}'
        return render_to_string(request, self.template, {'action': self})


class Link(PageAction):
    """
    A link action.

    Can point to anywhere.
    """

    template = 'ohmyadmin/actions/link.html'

    def __init__(self, label: str, url: str | LazyURL, icon: str = '', variant: ButtonType = 'link') -> None:
        self.url = url
        self.icon = icon
        self.label = label
        self.variant = variant
        self.slug = '-----'

    def resolve(self, request: Request) -> URL:
        return resolve_url(request, self.url)


class Submit(PageAction):
    """
    Submit actions are regular buttons.

    Mostly for use with forms.
    """

    template = 'ohmyadmin/actions/submit.html'

    def __init__(
        self,
        label: str,
        variant: ButtonType = 'default',
        name: str = '',
        html_attrs: typing.Mapping[str, typing.Any] | None = None,
        type: typing.Literal['button', 'submit'] = 'submit',
    ) -> None:
        self.name = name
        self.label = label
        self.type = type
        self.variant = variant
        self.html_attrs = html_attrs or {}


class Dispatch(abc.ABC):  # pragma: no cover
    slug: str

    @abc.abstractmethod
    async def dispatch(self, request: Request) -> Response:
        ...


class Callback(Dispatch, PageAction):
    """Callback actions execute a backend function when clicked."""

    template = 'ohmyadmin/actions/callback.html'

    def __init__(
        self,
        slug: str,
        label: str,
        callback: ActionCallback,
        icon: str = '',
        variant: ButtonType = 'default',
        confirmation: str = '',
        http_method: RequestMethod = 'get',
        hx_target: str = '',
    ) -> None:
        self.icon = icon
        self.label = label
        self.http_method = http_method
        self.callback = callback
        self.hx_target = hx_target
        self.variant = variant
        self.confirmation = confirmation
        self.slug = slug

    def render(self, request: Request) -> str:
        menu_link = request.url.include_query_params(_action=self.slug)
        return render_to_string(request, self.template, {'action': self, 'menu_link': menu_link})

    async def dispatch(self, request: Request) -> Response:
        return await self.callback(request)


class ObjectAction(abc.ABC):
    template: str = ''

    def render(self, request: Request, obj: typing.Any) -> str:
        assert self.template, f'template is not defined on class {self.__class__.__name__}'
        return render_to_string(request, self.template, {'action': self, 'object': obj})


class ObjectLink(ObjectAction):
    """Renders as a dropdown menu item."""

    template = 'ohmyadmin/actions/object_link.html'

    def __init__(self, label: str, url: str | LazyURL | ObjectURLFactory, icon: str = '') -> None:
        self.url = url
        self.icon = icon
        self.label = label

    def resolve(self, request: Request, obj: typing.Any) -> URL:
        if isinstance(self.url, (str, LazyURL)):
            return resolve_url(request, self.url)
        return URL(str(self.url(request, obj)))


class ObjectCallback(Dispatch, ObjectAction):
    """
    Renders as dropdown menu item.

    When clicked, dispatches AJAX call to API.
    """

    slug: str = ''
    template = 'ohmyadmin/actions/object_callback.html'

    def __init__(
        self,
        slug: str,
        label: str,
        callback: ActionCallback,
        icon: str = '',
        dangerous: bool = False,
        http_method: RequestMethod = 'get',
        confirmation: str = '',
    ) -> None:
        self.slug = slug
        self.icon = icon
        self.label = label
        self.http_method = http_method
        self.callback = callback
        self.dangerous = dangerous
        self.confirmation = confirmation

    def resolve_url(self, request: Request, obj: typing.Any) -> URL:
        params = MultiDict(parse_qsl(request.url.query, keep_blank_values=True))
        params.append('_object_action', self.slug)
        params.setlist('_ids', [request.state.datasource.get_pk(obj)])
        return request.url.replace(query=urlencode(params.multi_items()))

    async def dispatch(self, request: Request) -> Response:
        return await self.callback(request)


class BasePageAction(Callback):
    icon: str = ''
    button_color: ButtonType = 'default'
    label: str = '<unlabeled>'
    confirmation: str = ''
    method: RequestMethod = 'get'
    slug: str = ''

    def __init__(self, slug: str = '') -> None:
        slug = slug or self.slug or slugify(self.__class__.__name__.removesuffix('PageAction'))
        super().__init__(
            slug,
            label=self.label,
            callback=self.dispatch,
            icon=self.icon,
            variant=self.button_color,
            confirmation=self.confirmation,
            http_method=self.method,
            hx_target='#modals',
        )

    @abc.abstractmethod
    async def apply(self, request: Request) -> Response:  # pragma: no cover
        ...

    async def dispatch(self, request: Request) -> Response:
        return await self.apply(request)


class BaseFormPageAction(BasePageAction):
    form_class: type[wtforms.Form] = wtforms.Form
    actions: list[Submit | Link] | None = None
    form_template: str = 'ohmyadmin/actions/form_action.html'

    async def get_form_object(self, request: Request) -> typing.Any:
        return None

    def get_form_actions(self, request: Request, model: typing.Any) -> list[Submit | Link]:
        return self.actions or [
            Submit(label=_('Submit', domain='ohmyadmin'), variant='accent', type='submit'),
            Submit(
                label=_('Cancel', domain='ohmyadmin'),
                variant='text',
                type='button',
                html_attrs={
                    '@click': 'closeModal;',
                },
            ),
        ]

    async def apply(self, request: Request) -> Response:
        model = await self.get_form_object(request)
        form = await create_form(request, self.form_class, model)
        if await validate_on_submit(request, form):
            return await self.handle(request, form, model)

        return render_to_response(
            request,
            self.form_template,
            {'action': self, 'form': form, 'actions': self.get_form_actions(request, model)},
        )

    @abc.abstractmethod
    async def handle(self, request: Request, form: wtforms.Form, model: typing.Any) -> Response:  # pragma: no cover
        ...


class BaseObjectAction(ObjectCallback):
    icon: str = ''
    label: str = '<unlabeled>'
    dangerous: bool = False
    confirmation: str = ''
    http_method: RequestMethod = 'get'

    def __init__(self, slug: str = '') -> None:
        slug = slug or self.slug or slugify(self.__class__.__name__.removesuffix('ObjectAction'))
        super().__init__(
            slug,
            label=self.label,
            callback=self.dispatch,
            icon=self.icon,
            dangerous=self.dangerous,
            http_method=self.http_method,
            confirmation=self.confirmation,
        )

    @abc.abstractmethod
    async def apply(self, request: Request, object_id: str) -> Response:  # pragma: no cover
        ...

    async def dispatch(self, request: Request) -> Response:
        object_id = request.query_params.get('_ids', '')
        if not object_id:
            return ActionResponse().show_toast(_('No object selected.', domain='ohmyadmin'), 'error')
        return await self.apply(request, object_id)


class BaseFormObjectAction(BaseObjectAction):
    form_class: type[wtforms.Form] = wtforms.Form
    actions: list[Submit | Link] | None = None
    form_template: str = 'ohmyadmin/actions/form_action.html'

    async def get_form_object(self, request: Request, object_id: str) -> typing.Any:
        return None

    def get_form_actions(self, request: Request, model: typing.Any) -> list[Submit | Link]:
        return self.actions or [
            Submit(label=_('Submit', domain='ohmyadmin'), variant='accent', type='submit'),
            Submit(
                label=_('Cancel', domain='ohmyadmin'),
                variant='text',
                type='button',
                html_attrs={
                    '@click': 'closeModal;',
                },
            ),
        ]

    async def apply(self, request: Request, object_id: str) -> Response:
        model = await self.get_form_object(request, object_id)
        form = await create_form(request, self.form_class, model)
        if await validate_on_submit(request, form):
            return await self.handle(request, form, model)

        return render_to_response(
            request,
            self.form_template,
            {'action': self, 'form': form, 'actions': self.get_form_actions(request, model)},
        )

    @abc.abstractmethod
    async def handle(self, request: Request, form: wtforms.Form, model: typing.Any) -> Response:  # pragma: no cover
        ...


class BatchAction(ObjectCallback, abc.ABC):
    slug: str = ''

    @abc.abstractmethod
    async def apply(self, request: Request, object_ids: list[str], form: wtforms.Form) -> Response:  # pragma: no cover
        ...


class BaseBatchAction(BatchAction):
    label: str = '<unlabeled>'
    dangerous: bool = False
    message: str = _('Do you really want to appy this action on selected rows?', domain='ohmyadmin')
    form_class: type[wtforms.Form] = wtforms.Form
    template = 'ohmyadmin/actions/batch_action_modal.html'

    def __init__(self, slug: str = '') -> None:
        slug = slug or self.slug or slugify(self.__class__.__name__.removesuffix('BatchAction'))
        super().__init__(
            slug=slug, label=self.label, callback=self.dispatch, dangerous=self.dangerous, http_method='get'
        )

    async def dispatch(self, request: Request) -> Response:
        object_ids = request.query_params.getlist('_ids')
        form = await create_form(request, self.form_class)
        if await validate_on_submit(request, form):
            return await self.apply(request, object_ids, form)
        return render_to_response(request, self.template, {'object_ids': object_ids, 'action': self, 'form': form})


class BatchDelete(BaseBatchAction):
    dangerous = True
    success_message = _('{count} records has been deleted.', domain='ohmyadmin')
    message = _('Are you sure you want to delete selected objects?', domain='ohmyadmin')
    label = _('Mass delete', domain='ohmyadmin')

    async def apply(self, request: Request, object_ids: list[str], form: wtforms.Form) -> Response:
        datasource: DataSource = request.state.datasource
        await datasource.delete(request, *object_ids)
        return ActionResponse().show_toast(self.success_message.format(count=len(object_ids))).close_modal()


class DeleteObjectAction(BaseObjectAction):
    icon = 'trash'
    dangerous = True
    label = _('Delete', domain='ohmyadmin')
    success_message = _('{object} has been deleted.', domain='ohmyadmin')
    confirmation = _('Are you sure you want to delete this record?', domain='ohmyadmin')
    http_method = 'delete'

    async def apply(self, request: Request, object_id: str) -> Response:
        if request.method in ['POST', 'DELETE']:
            datasource: DataSource = request.state.datasource
            model = await datasource.get(request, object_id)
            await datasource.delete(request, object_id)
            return ActionResponse().show_toast(self.success_message.format(object=model)).refresh_datatable()
        return ActionResponse().show_toast(_('Unsupported HTTP method.'), 'error')



import inspect
import typing
from starlette.concurrency import run_in_threadpool
from starlette.exceptions import HTTPException
from starlette.requests import Request
from starlette.responses import Response
from starlette.routing import BaseRoute, Route
from starlette.types import Receive, Scope, Send

from ohmyadmin.pages.base import BasePage


class Page(BasePage):
    """
    A simple page that renders a template with context. Much like Starlette's HTTPEndpoint class.

    By default, it generates `get` method that returns template defined by `template` class attribute.

    ## Example:

    The typical usage is like this:
    ```python title="myapp/demo_page.py"
    from ohmyadmin.pages.page import Page

    class MyPage(Page):
        template = 'myproject/page.html

        async def handle(self, request: Request) -> dict[str, typing.Any]:
            # return dict of variables for the template
            return {
                'hello': 'world',
            }
    ```
    """

    template: typing.ClassVar[str] = 'ohmyadmin/pages/blank.html'

    async def handle(self, request: Request) -> typing.Mapping[str, typing.Any]:
        """Execute business logic here and return template context."""
        return {}

    async def get(self, request: Request) -> Response:
        context = dict(await self.handle(request))
        context.update({'page_title': self.page_title})
        return self.render_to_response(request, self.template, context)

    async def dispatch(self, request: Request) -> Response:
        method = request.method.lower()
        if handler := getattr(self, method, None):
            return (
                await handler(request)
                if inspect.iscoroutinefunction(handler)
                else await run_in_threadpool(handler, request)
            )
        raise HTTPException(405, 'Method Not Allowed')  # pragma: no cover

    def as_route(self) -> BaseRoute:
        """
        Create route for this page.

        It will automatically detect used HTTP methods using defined methods. For example, if page has `def get` and
        `def post` methods then route will serve GET, HEAD, and POST verbs.
        """
        methods: list[str] = []
        for method in ['GET', 'POST', 'PUT', 'PATCH', 'DELETE']:
            if hasattr(self, method.lower()):
                methods.append(method)

        return Route(f'/{self.slug}', self, methods=methods, name=self.get_path_name())

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        """ASGI integration point."""
        request = Request(scope, receive, send)
        response = await self.dispatch(request)
        await response(scope, receive, send)



import typing
from slugify import slugify
from starlette.datastructures import URL
from starlette.requests import Request
from starlette.responses import RedirectResponse, Response
from starlette.routing import BaseRoute

from ohmyadmin.helpers import camel_to_sentence, pluralize


class PageMeta(type):
    def __new__(mcs, name: str, bases: tuple[type], attrs: dict[str, typing.Any], **kwargs: typing.Any) -> type:
        if '__abstract__' not in attrs:
            if not attrs.get('label'):
                clean_label = name.removesuffix('Page').removesuffix('Resource')
                attrs['label'] = camel_to_sentence(clean_label)
            if not attrs.get('label_plural'):
                attrs['label_plural'] = attrs.get('label_plural', pluralize(attrs['label']))
            if not attrs.get('slug'):
                attrs['slug'] = slugify(attrs['label'])
            if not attrs.get('page_title'):
                attrs['page_title'] = attrs['label']

        return type.__new__(mcs, name, bases, attrs, **kwargs)


class BasePage(metaclass=PageMeta):
    slug: str = ''
    label: str = ''
    page_title: str = ''
    label_plural: str = ''
    group: str = ''
    icon: str = ''

    def render_to_response(
        self,
        request: Request,
        template_name: str,
        context: typing.Mapping[str, typing.Any] | None = None,
        headers: typing.Mapping[str, typing.Any] | None = None,
    ) -> Response:
        """Render template into HTTP response."""
        return request.state.admin.render_to_response(request, template_name, context, headers=headers)

    def redirect_to_path(self, request: Request, path_name: str, **path_params: typing.Any) -> RedirectResponse:
        """A shortcut to create RedirectResponse for path name and params."""
        url = request.url_for(path_name, **path_params)
        return RedirectResponse(url, status_code=302)

    def redirect_to(self, url: str | URL) -> RedirectResponse:
        """A shortcut to create RedirectResponse for given URL."""
        return RedirectResponse(url, status_code=302)

    def redirect_to_self(self, request: Request) -> RedirectResponse:
        """A shortcut to create RedirectResponse back to this page."""
        url = request.url_for(self.get_path_name())
        return RedirectResponse(url, status_code=302)

    def as_route(self) -> BaseRoute:  # pragma: no cover
        """Create a route instance for this page."""
        raise NotImplementedError()

    @classmethod
    def get_path_name(cls) -> str:
        """Generate a path name for this page."""
        return f'app.pages.{cls.slug}'

    @classmethod
    def generate_url(cls, request: Request) -> str:
        """Generate absolute URL to this page."""
        return request.url_for(cls.get_path_name())



import abc
import typing
import wtforms
from starlette.requests import Request
from starlette.responses import Response
from starlette.routing import BaseRoute, Route
from starlette_babel import gettext_lazy as _

from ohmyadmin import actions, layouts
from ohmyadmin.forms import create_form, validate_on_submit
from ohmyadmin.pages.base import BasePage

_F = typing.TypeVar('_F', bound=wtforms.Form)


class FormPage(BasePage):
    """This page renders HTML form."""

    __abstract__ = True

    form_class: type[wtforms.Form]
    form_actions: typing.Sequence[actions.Submit | actions.Link] | None = None
    template: typing.ClassVar[str] = 'ohmyadmin/pages/form.html'

    def get_form_actions(self, request: Request) -> typing.Sequence[actions.Submit | actions.Link]:
        return self.form_actions or [
            actions.Submit(label=_('Submit', domain='ohmyadmin'), variant='accent'),
        ]

    async def get_form_object(self, request: Request) -> typing.Any:
        return None

    async def create_form(self, request: Request, model: typing.Any) -> wtforms.Form:
        return await create_form(request, self.form_class, obj=model)

    def build_form_layout(self, request: Request, form: wtforms.Form) -> layouts.Layout:
        return layouts.Card([layouts.StackedForm([layouts.Input(field) for field in form])])

    async def dispatch(self, request: Request) -> Response:
        model = await self.get_form_object(request)
        form = await self.create_form(request, model)
        if await validate_on_submit(request, form):
            return await self.handle_submit(request, form, model)

        form_layout = self.build_form_layout(request, form)
        form_actions = self.get_form_actions(request)
        return self.render_to_response(
            request,
            self.template,
            {'form': form, 'page_title': self.page_title, 'form_layout': form_layout, 'form_actions': form_actions},
        )

    @abc.abstractmethod
    async def handle_submit(
        self, request: Request, form: wtforms.Form, model: typing.Any
    ) -> Response:  # pragma: no cover
        ...

    def as_route(self) -> BaseRoute:
        return Route(f'/{self.slug}', self.dispatch, methods=['GET', 'POST'], name=self.get_path_name())



from __future__ import annotations

import dataclasses

import abc
import decimal
import functools
import inspect
import typing
import wtforms
from starlette.requests import Request
from starlette_babel import gettext_lazy as _

from ohmyadmin.datasource.base import DataSource, NumberOperation, StringOperation
from ohmyadmin.forms import AsyncSelectField, AsyncSelectMultipleField, init_form
from ohmyadmin.helpers import snake_to_sentence
from ohmyadmin.shortcuts import render_to_string

_FT = typing.TypeVar('_FT', bound=wtforms.Form)
_T = typing.TypeVar('_T', bound='BaseFilter')


class BaseFilter(abc.ABC, typing.Generic[_FT]):
    """Base filter class for data filters."""

    template: str = 'ohmyadmin/filters/form.html'
    indicator_template: str = 'ohmyadmin/filters/blank_indicator.html'
    form_class: type[_FT] = wtforms.Form
    form: _FT

    def __new__(  # type: ignore[misc]
        cls,
        *args: typing.Any,
        **kwargs: typing.Any,
    ) -> UnboundFilter[_T] | BaseFilter[_FT]:
        """
        We attach filters as instances to let users customize it via constructor arguments.

        However, this creates mutable singletons. To workaround, we create unbound filters which are like factories and
        create actual instances when called. There may be a better idea but the current one is good enough for this
        purpose.
        """
        if '_create' in kwargs:
            return super().__new__(cls)

        return UnboundFilter(cls, args, kwargs)  # type: ignore[arg-type]

    def __init__(self, query_param: str, label: str = '', **kwargs: typing.Any) -> None:
        self.query_param = query_param
        self.label = label or snake_to_sentence(self.query_param.title())
        self.form = self.form_class(prefix=self.query_param)

    async def prepare(self, request: Request) -> None:
        """
        Creation hook to execute async operations when page creates the filter.

        As a user you have to override `BaseFilter.initialize` instead of this method.
        """
        self.form.process(request.query_params)
        await self.initialize(request)

    async def initialize(self, request: Request) -> None:
        """Override this method if you want to call async code when page instantiates the filter."""

    @abc.abstractmethod
    def apply(self, request: Request, query: DataSource) -> DataSource:
        """Apply filter to the data source query."""

    @abc.abstractmethod
    def is_active(self, request: Request) -> bool:
        """
        Check if the filter is active.

        Active filters rendered differently in the filter bar.
        """

    def get_indicator_context(self, value: dict[str, typing.Any]) -> dict[str, typing.Any]:
        """
        Various filters render different data in the indicators. This method returns template context for indicators.
        The context available in the indicator template via `indicator` template variable.

        :param value: current form data.
        """
        return value

    def render_form(self, request: Request) -> str:
        return render_to_string(request, self.template, {'filter': self, 'form': self.form})

    def render_indicator(self, request: Request) -> str:
        """
        Render indicator.

        Indicators visible only when filter is active.
        """
        clear_url = request.url.remove_query_params([form_field.name for form_field in self.form]).include_query_params(
            clear=1
        )
        indicator = self.get_indicator_context(self.form.data)
        return render_to_string(
            request,
            self.indicator_template,
            {
                'filter': self,
                'indicator': indicator,
                'clear_url': clear_url,
            },
        )


@dataclasses.dataclass
class UnboundFilter(typing.Generic[_T]):
    """A filter factory."""

    filter_class: type[_T]
    args: tuple[typing.Any, ...]
    kwargs: dict[str, typing.Any]

    async def create(self, request: Request) -> _T:
        """Create a filter instance."""
        instance = self.filter_class(*self.args, _create=True, **self.kwargs)
        await instance.prepare(request)
        return instance


def _safe_enum_coerce(value: typing.Any, choices: type[StringOperation | NumberOperation]) -> typing.Any:
    try:
        return choices[value]
    except KeyError:
        return None


class StringFilterForm(wtforms.Form):
    operation = wtforms.SelectField(
        choices=StringOperation.choices(),
        coerce=functools.partial(_safe_enum_coerce, choices=StringOperation),
    )
    query = wtforms.StringField(validators=[wtforms.validators.data_required()])


class StringFilter(BaseFilter[StringFilterForm]):
    """
    String filters let users query data that matches text and rule.

    Rules defined by StringOperation and are like `starts with`, `ends with`, etc.
    """

    form_class = StringFilterForm
    indicator_template = 'ohmyadmin/filters/enum_indicator.html'

    def apply(self, request: Request, query: DataSource) -> DataSource:
        operation = self.form.data['operation']
        if not operation:
            return query

        value = self.form.data['query']
        return query.apply_string_filter(self.query_param, operation, value)

    def is_active(self, request: Request) -> bool:
        return bool(self.form.data['query'])


class IntegerFilterForm(wtforms.Form):
    operation = wtforms.SelectField(
        choices=NumberOperation.choices(),
        coerce=functools.partial(_safe_enum_coerce, choices=NumberOperation),
    )
    query = wtforms.IntegerField(validators=[wtforms.validators.data_required()])


class IntegerFilter(BaseFilter[IntegerFilterForm]):
    form_class = IntegerFilterForm
    indicator_template = 'ohmyadmin/filters/enum_indicator.html'

    def apply(self, request: Request, query: DataSource) -> DataSource:
        operation = self.form.data['operation']
        if not operation:
            return query

        value = self.form.data['query']
        return query.apply_number_filter(self.query_param, operation, value)

    def is_active(self, request: Request) -> bool:
        return bool(self.form.data['query'])


class FloatFilterForm(wtforms.Form):
    operation = wtforms.SelectField(
        choices=NumberOperation.choices(),
        coerce=functools.partial(_safe_enum_coerce, choices=NumberOperation),
    )
    query = wtforms.FloatField(validators=[wtforms.validators.data_required()])


class FloatFilter(IntegerFilter):
    form_class = FloatFilterForm


class DecimalFilterForm(wtforms.Form):
    operation = wtforms.SelectField(
        choices=NumberOperation.choices(),
        coerce=functools.partial(_safe_enum_coerce, choices=NumberOperation),
    )
    query = wtforms.DecimalField(validators=[wtforms.validators.data_required()])


class DecimalFilter(IntegerFilter):
    form_class = DecimalFilterForm


class DateFilterForm(wtforms.Form):
    query = wtforms.DateField()


class DateFilter(BaseFilter[DateFilterForm]):
    form_class = DateFilterForm
    indicator_template = 'ohmyadmin/filters/date_indicator.html'

    def apply(self, request: Request, query: DataSource) -> DataSource:
        value = self.form.data['query']
        if not value:
            return query
        return query.apply_date_filter(self.query_param, value)

    def is_active(self, request: Request) -> bool:
        return bool(self.form.data['query'])


class DateRangeFilterForm(wtforms.Form):
    after = wtforms.DateField()
    before = wtforms.DateField()


class DateRangeFilter(BaseFilter[DateRangeFilterForm]):
    form_class = DateRangeFilterForm
    indicator_template = 'ohmyadmin/filters/date_range_indicator.html'

    def apply(self, request: Request, query: DataSource) -> DataSource:
        before = self.form.data['before']
        after = self.form.data['after']
        if before or after:
            return query.apply_date_range_filter(self.query_param, before, after)
        return query

    def is_active(self, request: Request) -> bool:
        return self.form.data['before'] or self.form.data['after']

    def get_indicator_context(self, value: dict[str, typing.Any]) -> dict[str, typing.Any]:
        value['format'] = 'date'
        return super().get_indicator_context(value)


class DateTimeRangeFilterForm(wtforms.Form):
    after = wtforms.DateTimeLocalField()
    before = wtforms.DateTimeLocalField()


class DateTimeRangeFilter(BaseFilter[DateRangeFilterForm]):
    form_class = DateTimeRangeFilterForm
    indicator_template = 'ohmyadmin/filters/date_range_indicator.html'

    def apply(self, request: Request, query: DataSource) -> DataSource:
        before = self.form.data['before']
        after = self.form.data['after']
        if self.is_active(request):
            return query.apply_date_range_filter(self.query_param, before, after)
        return query

    def is_active(self, request: Request) -> bool:
        return bool(self.form.data['before'] or self.form.data['after'])

    def get_indicator_context(self, value: dict[str, typing.Any]) -> dict[str, typing.Any]:
        value['format'] = 'datetime'
        return super().get_indicator_context(value)


class ChoiceFilterForm(wtforms.Form):
    choice = AsyncSelectField(
        label=_('Choices', domain='ohmyadmin'),
        validators=[wtforms.validators.data_required()],
        choices=[],
    )


class ChoiceFilter(BaseFilter[ChoiceFilterForm]):
    form_class = ChoiceFilterForm
    indicator_template = 'ohmyadmin/filters/choice_indicator.html'

    def __init__(
        self,
        query_param: str,
        label: str = '',
        *,
        choices: typing.Any,
        coerce: type[str | int | float | decimal.Decimal] = str,
        **kwargs: typing.Any,
    ) -> None:
        super().__init__(query_param, label, **kwargs)
        self.coerce = coerce
        self.form.choice.coerce = self.coerce
        if inspect.iscoroutinefunction(choices):
            self.form.choice.choices_loader = choices
        else:
            self.form.choice.choices = choices

    async def initialize(self, request: Request) -> None:
        await init_form(request, self.form)

    def apply(self, request: Request, query: DataSource) -> DataSource:
        if self.is_active(request):
            choice = self.form.data['choice']
            return query.apply_choice_filter(self.query_param, [choice], self.coerce)
        return query

    def is_active(self, request: Request) -> bool:
        if self.form.validate():
            return bool(self.form.data['choice'])
        return False

    def get_indicator_context(self, value: dict[str, typing.Any]) -> dict[str, typing.Any]:
        by_key = {choice[0]: choice[1] for choice in self.form.choice.choices}
        choice = by_key.get(value['choice'], '')
        return {'value': choice}


class MultiChoiceFilterForm(wtforms.Form):
    choice = AsyncSelectMultipleField(
        label=_('Select multiple', domain='ohmyadmin'),
        validators=[wtforms.validators.data_required()],
        choices=[],
    )


class MultiChoiceFilter(BaseFilter[MultiChoiceFilterForm]):
    form_class = MultiChoiceFilterForm
    indicator_template = 'ohmyadmin/filters/multi_choice_indicator.html'

    def __init__(
        self,
        query_param: str,
        label: str = '',
        *,
        choices: typing.Any,
        coerce: type[str | int | float | decimal.Decimal] = str,
        **kwargs: typing.Any,
    ) -> None:
        super().__init__(query_param, label, **kwargs)
        self.coerce = coerce
        if inspect.iscoroutinefunction(choices):
            self.form.choice.choices_loader = choices
        else:
            self.form.choice.choices = choices
        self.form.choice.coerce = self.coerce

    async def initialize(self, request: Request) -> None:
        await init_form(request, self.form)

    def apply(self, request: Request, query: DataSource) -> DataSource:
        if self.is_active(request):
            choice = self.form.data['choice']
            return query.apply_choice_filter(self.query_param, choice, self.coerce)
        return query

    def is_active(self, request: Request) -> bool:
        if self.form.validate():
            return bool(self.form.data['choice'])
        return False

    def get_indicator_context(self, value: dict[str, typing.Any]) -> dict[str, typing.Any]:
        choices = (choice for choice in self.form.choice.choices if choice[0] in value['choice'])
        values = [choice[1] for choice in choices]
        return {'value': values}



from starlette.requests import HTTPConnection, Request
from starlette.responses import RedirectResponse
from starlette.types import ASGIApp, Receive, Scope, Send
from starlette_babel import gettext_lazy as _
from starlette_flash import flash


class LoginRequiredMiddleware:
    def __init__(self, app: ASGIApp, exclude_paths: list[str]) -> None:
        self.app = app
        self.exclude_paths = exclude_paths

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        if scope["type"] not in ["http", "websocket"]:
            await self.app(scope, receive, send)
            return

        conn = HTTPConnection(scope)
        if conn.user.is_authenticated or any([excluded_path in conn.url.path for excluded_path in self.exclude_paths]):
            await self.app(scope, receive, send)
            return

        if scope['type'] == 'http':
            flash(Request(scope)).error(_('You need to be logged in to access this page.', domain='ohmyadmin'))

        redirect_to = conn.url_for('ohmyadmin.login') + '?next=' + conn.url.path
        response = RedirectResponse(url=redirect_to, status_code=302)
        await response(scope, receive, send)



import dataclasses

import typing
from starlette.datastructures import URL, MultiDict
from starlette.requests import Request
from urllib.parse import parse_qsl, urlencode

SortingType = typing.Literal['asc', 'desc']


@dataclasses.dataclass
class Ordering:
    field: str
    direction: SortingType
    next_url: URL


def get_ordering_value(request: Request, param_name: str) -> dict[str, SortingType]:
    """
    Extract ordering value from query params.

    Returns a dictionary where keys are parameter names and values either 'asc' or 'desc' literals.
    """
    return {
        value[1:] if value.startswith('-') else value: 'desc' if value.startswith('-') else 'asc'
        for value in request.query_params.getlist(param_name)
        if value
    }


@dataclasses.dataclass
class SortControl:
    index: int | None
    url: URL
    show_index: bool
    ordering: SortingType | None


class SortingHelper:
    """API interface for managing ordering values from query parameters."""

    def __init__(self, request: Request, query_param: str) -> None:
        self.request = request
        self.query_param_name = query_param
        self.ordering = request.query_params.getlist(query_param)

    def get_current_ordering(self, field: str) -> SortingType | None:
        for order in self.ordering:
            if order == field:
                return 'asc'
            if order == f'-{field}':
                return 'desc'

        return None

    def get_current_ordering_index(self, field: str) -> int | None:
        for index, param_name in enumerate(self.ordering):
            if param_name.endswith(field):
                return index + 1
        return None

    def get_url(self, field: str) -> URL:
        """
        Generate a URL with inverted ordering params.

        For example, if current page has `ordering=title` then the function generates `ordering=-title`.
        """
        ordering = self.ordering.copy()
        if field in ordering:
            index = ordering.index(field)
            ordering[index] = f'-{field}'
        elif f'-{field}' in ordering:
            ordering.remove(f'-{field}')
        else:
            ordering.append(field)

        params = MultiDict(parse_qsl(self.request.url.query, keep_blank_values=True))
        params.setlist(self.query_param_name, ordering)
        url = self.request.url.replace(query=urlencode(params.multi_items()))
        return url

    def should_show_index(self) -> bool:
        """
        Test if current ordering index should be displayed next to column label.

        Typically used to highlight in what order the data source has been sorted.
        """
        return len(self.ordering) > 1

    def get_control(self, field: str) -> SortControl:
        url = self.get_url(field)
        ordering = self.get_current_ordering(field)
        index = self.get_current_ordering_index(field)
        show_index = self.should_show_index()
        return SortControl(index=index, ordering=ordering, url=url, show_index=show_index)



import jinja2
import typing
from markupsafe import Markup


def as_html_attrs(value: typing.Mapping) -> str:
    def resolve_attr_name(name: str) -> str:
        if name.startswith('data') or name.startswith('aria'):
            return name.replace('_', '-')
        return name

    parts: list[str] = []
    for attr_name, attr_value in value.items():
        attr_name = resolve_attr_name(attr_name)
        if not attr_value:
            continue
        if attr_value is True:
            parts.append(attr_name)
            continue
        parts.append(f'{attr_name}="{attr_value}"')
    return Markup(' '.join(parts))


@jinja2.pass_context
def pk_filter(context: jinja2.runtime.Context, obj: typing.Any) -> str:
    return context['request'].state.datasource.get_pk(obj)



import json
import typing
from starlette.requests import Request
from starlette.responses import Response
from starlette.routing import BaseRoute, Mount, Route
from starlette_babel import gettext_lazy as _
from unittest import mock

from ohmyadmin import actions
from ohmyadmin.datasource.base import DataSource
from ohmyadmin.filters import BaseFilter, UnboundFilter
from ohmyadmin.metrics import Card, UndefinedCardError
from ohmyadmin.ordering import get_ordering_value
from ohmyadmin.pages.page import Page
from ohmyadmin.pagination import Pagination, get_page_size_value, get_page_value
from ohmyadmin.shortcuts import render_to_response
from ohmyadmin.views.base import IndexView
from ohmyadmin.views.table import TableColumn, TableView


def get_search_value(request: Request, param_name: str) -> str:
    return request.query_params.get(param_name, '').strip()


class TablePage(Page):
    """Table pages display lists of objects with option to sort, search, and filter."""

    datasource: DataSource
    page_param: typing.ClassVar[str] = 'page'
    page_size_param: typing.ClassVar[str] = 'page_size'
    search_param: typing.ClassVar[str] = 'search'
    ordering_param: typing.ClassVar[str] = 'ordering'
    page_size: typing.ClassVar[int] = 25
    max_page_size: typing.ClassVar[int] = 100
    columns: typing.Sequence[TableColumn] | None = None

    metrics: typing.Sequence[type[Card]] | None = None
    page_actions: typing.Sequence[actions.PageAction] | None = None
    filters: typing.Sequence[UnboundFilter | BaseFilter] | None = None
    object_actions: typing.Sequence[actions.ObjectAction] | None = None
    batch_actions: typing.Sequence[actions.BatchAction] | None = None

    __abstract__ = True

    def get_sortable_fields(self, request: Request) -> list[str]:
        return [column.sort_by for column in self.get_table_columns(request) if column.sortable]

    def get_searchable_fields(self, request: Request) -> list[str]:
        return [column.search_in for column in self.get_table_columns(request) if column.searchable]

    def is_searchable(self, request: Request) -> bool:
        return bool(self.get_searchable_fields(request))

    def get_search_placeholder(self, request: Request) -> str:
        template = _('Search in {fields}.')
        fields = ', '.join([str(column.label) for column in self.get_table_columns(request) if column.searchable])
        return template.format(fields=fields)

    def get_table_columns(self, request: Request) -> typing.Sequence[TableColumn]:
        return self.columns or []

    def get_filters(self, request: Request) -> typing.Sequence[UnboundFilter]:
        return self.filters or []  # type: ignore[return-value]

    def get_page_actions(self, request: Request) -> list[actions.PageAction]:
        return list(self.page_actions or [])

    def get_object_actions(self, request: Request, obj: typing.Any) -> list[actions.ObjectAction]:
        return list(self.object_actions or [])

    def get_batch_actions(self, request: Request) -> list[actions.BatchAction]:
        return list(self.batch_actions or [])

    def get_metrics(self, request: Request) -> typing.Sequence[Card]:
        return [metric() for metric in self.metrics or []]

    async def dispatch_metric(self, request: Request, metric_slug: str) -> Response:
        metrics = {metric.slug: metric for metric in self.get_metrics(request)}
        try:
            metric = metrics[metric_slug]
        except KeyError:
            raise UndefinedCardError(f'Metric "{metric_slug}" is not defined.')
        return await metric.dispatch(request)

    async def dispatch_batch_action(self, request: Request, action_slug: str) -> Response:
        actions_ = {action.slug: action for action in self.get_batch_actions(request)}
        try:
            action = actions_[action_slug]
        except KeyError:
            raise actions.UndefinedActionError(f'Batch action "{action_slug}" is not defined.')

        return await action.dispatch(request)

    async def dispatch_object_action(self, request: Request, action_slug: str) -> Response:
        actions_ = {
            action.slug: action
            for action in self.get_object_actions(request, mock.MagicMock()) or []
            if isinstance(action, actions.Dispatch)
        }

        try:
            action = actions_[action_slug]
        except KeyError:
            raise actions.UndefinedActionError(f'Object action "{action_slug}" is not defined.')

        return await action.dispatch(request)

    async def dispatch_action(self, request: Request, action_slug: str) -> Response:
        actions_ = {
            action.slug: action for action in self.get_page_actions(request) if isinstance(action, actions.Dispatch)
        }

        try:
            action = actions_[action_slug]
        except KeyError:
            raise actions.UndefinedActionError(f'Action "{action_slug}" is not defined.')

        return await action.dispatch(request)

    async def create_filters(self, request: Request) -> list[BaseFilter]:
        return [await _filter.create(request) for _filter in self.get_filters(request)]

    async def get_objects(self, request: Request, filters: list[BaseFilter]) -> Pagination:
        page_number = get_page_value(request, self.page_param)
        page_size = get_page_size_value(request, self.page_size_param, self.max_page_size, self.page_size)
        search_term = get_search_value(request, self.search_param)
        ordering = get_ordering_value(request, self.ordering_param)

        query = self.datasource.get_query_for_index()
        if search_term:
            query = query.apply_search(search_term, self.get_searchable_fields(request))

        if ordering:
            query = query.apply_ordering(ordering, self.get_sortable_fields(request))

        for _filter in filters:
            query = _filter.apply(request, query)

        return await query.paginate(request, page=page_number, page_size=page_size)

    def as_route(self) -> BaseRoute:
        methods = ['GET', 'POST', 'PUT', 'PATCH', 'DELETE']
        return Mount(
            f'/{self.slug}',
            routes=[
                Route('/', self.dispatch, methods=methods, name=self.get_path_name()),
            ],
        )

    def get_view(self, request: Request) -> IndexView:
        return TableView(columns=self.get_table_columns(request))

    async def dispatch(self, request: Request) -> Response:
        request.state.page = self
        request.state.datasource = self.datasource

        if '_action' in request.query_params:
            return await self.dispatch_action(request, request.query_params['_action'])
        if '_object_action' in request.query_params:
            return await self.dispatch_object_action(request, request.query_params['_object_action'])
        if '_batch_action' in request.query_params:
            return await self.dispatch_batch_action(request, request.query_params['_batch_action'])
        if '_metric' in request.query_params:
            return await self.dispatch_metric(request, request.query_params['_metric'])

        filters = await self.create_filters(request)
        objects = await self.get_objects(request, filters)
        view = self.get_view(request)
        view_content = view.render(request, objects)

        if request.headers.get('hx-target', '') == 'filter-bar':
            headers = {}
            if 'clear' in request.query_params:
                headers = {
                    'hx-push-url': str(request.url.remove_query_params('clear')),
                    'hx-trigger-after-settle': json.dumps({'refresh-datatable': ''}),
                }
            return render_to_response(
                request,
                'ohmyadmin/pages/table/_filters.html',
                {'filters': filters},
                headers=headers,
            )

        if request.headers.get('hx-target', '') == 'data':
            return render_to_response(
                request,
                'ohmyadmin/pages/table/_content.html',
                {
                    'request': request,
                    'objects': objects,
                    'view_content': view_content,
                },
                headers={
                    'hx-push-url': str(request.url),
                    'hx-trigger-after-settle': json.dumps({'filters-reload': ''}),
                },
            )

        active_filter_count = sum([1 for filter in filters if filter.is_active(request)])
        return render_to_response(
            request,
            'ohmyadmin/pages/table/table.html',
            {
                'page': self,
                'objects': objects,
                'filters': filters,
                'view_content': view_content,
                'page_title': self.label_plural,  # type: ignore[attr-defined]
                'active_filter_count': active_filter_count,
                'search_term': get_search_value(request, self.search_param),
            },
        )



import dataclasses

import abc
import typing
from slugify import slugify
from starlette.datastructures import URL
from starlette.requests import Request
from starlette.responses import Response

from ohmyadmin.helpers import camel_to_sentence
from ohmyadmin.shortcuts import render_to_response


class CardError(Exception):
    ...


class UndefinedCardError(CardError):
    ...


class Card(abc.ABC):
    slug: str = ''
    label: str = ''
    size: int = 4

    def __init__(self) -> None:
        self.slug = self.slug or slugify(self.__class__.__name__)
        self.label = self.label or camel_to_sentence(self.__class__.__name__)

    @abc.abstractmethod
    async def dispatch(self, request: Request) -> Response:  # pragma: no cover
        ...

    def resolve_url(self, request: Request) -> URL:
        return request.url.include_query_params(_metric=self.slug)


class ValueMetric(Card):
    template: str = 'ohmyadmin/metrics/value.html'

    @abc.abstractmethod
    async def calculate(self, request: Request) -> str:  # pragma: no cover
        ...

    async def dispatch(self, request: Request) -> Response:
        value = await self.calculate(request)
        return render_to_response(request, self.template, {'value': value, 'metric': self})


@dataclasses.dataclass
class TrendResult:
    current_value: int | float | None = None
    series: list[tuple[str, float]] = dataclasses.field(default_factory=list)


class TrendMetric(Card):
    template: str = 'ohmyadmin/metrics/trend.html'

    @abc.abstractmethod
    async def calculate(self, request: Request) -> TrendResult:  # pragma: no cover
        ...

    async def dispatch(self, request: Request) -> Response:
        value = await self.calculate(request)
        return render_to_response(
            request,
            self.template,
            {'value': value, 'metric': self},
        )


class ProgressMetric(Card):
    """
    Renders progress bar metric.

    Useful for displaying the progresses.
    """

    target: int = 100
    progress_color: str = 'green'
    template: str = 'ohmyadmin/metrics/progress.html'

    async def get_target(self, request: Request) -> int:
        """
        The value you attempt to reach.

        Examples: 100 percents, or 1000 users
        """
        return self.target

    @abc.abstractmethod
    async def calculate(self, request: Request) -> float | int:  # pragma: no cover
        """
        Calculate the current value.

        This value is used against the target to calculate percentage.
        """

    async def dispatch(self, request: Request) -> Response:
        value = await self.calculate(request)
        target = await self.get_target(request)
        percent = f'{value * 100 / target:.0f}'
        return render_to_response(
            request,
            self.template,
            {
                'value': value,
                'target': target,
                'percent': percent,
                'metric': self,
            },
        )


class PartitionItem(typing.TypedDict):
    color: str
    value: float


@dataclasses.dataclass
class PartitionResult:
    groups: dict[str, PartitionItem] = dataclasses.field(default_factory=dict)

    def __iter__(self) -> typing.Iterator[tuple[str, PartitionItem]]:
        return iter(self.groups.items())


class PartitionMetric(Card):
    """Renders a pie chart."""

    template: str = 'ohmyadmin/metrics/partition.html'

    @abc.abstractmethod
    async def calculate(self, request: Request) -> PartitionResult:  # pragma: no cover
        ...

    async def dispatch(self, request: Request) -> Response:
        partitions = await self.calculate(request)
        return render_to_response(
            request,
            self.template,
            {
                'partitions': partitions,
                'metric': self,
            },
        )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['ohmyadmin',
 'ohmyadmin.contrib',
 'ohmyadmin.contrib.sqlalchemy',
 'ohmyadmin.datasource',
 'ohmyadmin.pages',
 'ohmyadmin.views']

package_data = \
{'': ['*'],
 'ohmyadmin': ['statics/.gitkeep',
               'templates/ohmyadmin/*',
               'templates/ohmyadmin/actions/*',
               'templates/ohmyadmin/filters/*',
               'templates/ohmyadmin/formatters/*',
               'templates/ohmyadmin/layouts/*',
               'templates/ohmyadmin/lib/*',
               'templates/ohmyadmin/metrics/*',
               'templates/ohmyadmin/pages/*',
               'templates/ohmyadmin/pages/table/*',
               'templates/ohmyadmin/resources/*',
               'templates/ohmyadmin/views/table/*']}

install_requires = \
['Jinja2>=3,<4',
 'WTForms>=3,<4',
 'async-storages>=0.4.1,<0.5.0',
 'python-multipart>=0.0.5,<0.0.6',
 'python-slugify>=7.0,<8.0',
 'starlette',
 'starlette-babel',
 'starlette-flash',
 'tabler-icons']

setup_kwargs = {
    'name': 'ohmyadmin',
    'version': '0.4.0',
    'description': 'Awesome admin panel for your business.',
    'long_description': '# OhMyAdmin\n\nAwesome admin panel for your business.\n\n![PyPI](https://img.shields.io/pypi/v/ohmyadmin)\n![GitHub Workflow Status](https://img.shields.io/github/workflow/status/alex-oleshkevich/ohmyadmin/Lint%20and%20test)\n![GitHub](https://img.shields.io/github/license/alex-oleshkevich/ohmyadmin)\n![Libraries.io dependency status for latest release](https://img.shields.io/librariesio/release/pypi/ohmyadmin)\n![PyPI - Downloads](https://img.shields.io/pypi/dm/ohmyadmin)\n![GitHub Release Date](https://img.shields.io/github/release-date/alex-oleshkevich/ohmyadmin)\n\n## Installation\n\nInstall `ohmyadmin` using PIP or poetry:\n\n```bash\npip install ohmyadmin\n# or\npoetry add ohmyadmin\n```\n\n## Features\n\n-   TODO\n\n## Quick start\n\nSee example application in `examples/` directory of this repository.\n',
    'author': 'Alex Oleshkevich',
    'author_email': 'alex.oleshkevich@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/alex-oleshkevich/ohmyadmin',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)



from __future__ import annotations

import datetime
import decimal
import sqlalchemy as sa
import typing
import uuid
from sqlalchemy import orm
from sqlalchemy.sql.elements import NamedColumn
from starlette.requests import Request

from ohmyadmin.datasource.base import DataSource, NumberOperation, StringOperation
from ohmyadmin.ordering import SortingType
from ohmyadmin.pagination import Pagination


def get_column_properties(model_class: typing.Any, prop_names: typing.Sequence[str]) -> dict[str, orm.ColumnProperty]:
    """
    Return SQLAlchemy columns defined on entity class by their string names.

    Looks up in the relations too.
    """
    mapper: orm.Mapper = orm.class_mapper(model_class)
    props: dict[str, orm.ColumnProperty] = {}
    for name in prop_names:
        if name in mapper.all_orm_descriptors:
            props[name] = mapper.all_orm_descriptors[name].property  # type: ignore[attr-defined]
        elif '.' in name:
            related_attr, related_column = name.split('.')
            if related_property := mapper.all_orm_descriptors.get(related_attr):
                props[name] = related_property.entity.all_orm_descriptors[  # type: ignore[attr-defined]
                    related_column
                ].property  # type: ignore[attr-defined]
    return props


def create_search_token(column: NamedColumn, search_query: str) -> sa.sql.ColumnElement:
    string_column = sa.cast(column, sa.Text)
    if search_query.startswith('^'):
        search_token = f'{search_query[1:].lower()}%'
        return string_column.startswith(search_token)

    if search_query.startswith('='):
        search_token = f'{search_query[1:].lower()}'
        return string_column == search_token

    if search_query.startswith('@'):
        search_token = f'{search_query[1:].lower()}'
        return string_column.regexp_match(search_token)

    if search_query.startswith('$'):
        search_token = f'{search_query[1:].lower()}'
        return string_column.endswith(search_token)

    search_token = f'%{search_query.lower()}%'
    return string_column.ilike(search_token)


def guess_pk_field(model_class: type) -> str:
    mapper: orm.Mapper = orm.class_mapper(model_class)
    pk_columns = [
        c.name for c in mapper.all_orm_descriptors if hasattr(c, 'primary_key') and c.primary_key  # type: ignore
    ]
    assert len(pk_columns), f'Model class {model_class} does not contain any primary key.'
    assert len(pk_columns) == 1, f'Model class {model_class} defines composite primary key which are not supported.'
    return pk_columns[0]


def guess_pk_type(model_class: type, pk_field: str) -> typing.Callable:
    mapper: orm.Mapper = orm.class_mapper(model_class)
    column = mapper.columns[pk_field]
    match column.type:
        case sa.String() | sa.Text():
            return str
        case sa.Integer():
            return int
        case sa.Float():
            return float
        case sa.Numeric():
            return decimal.Decimal
        case sa.Uuid():
            return uuid.UUID
    raise ValueError(f'Failed to guess primary key column caster for {column} (type={column.type}).')


T = typing.TypeVar('T', bound=typing.Any)


class SQLADataSource(DataSource[T]):
    def __init__(
        self,
        model_class: type[T],
        query: sa.Select | None = None,
        query_for_list: sa.Select | None = None,
        pk_column: str | None = None,
        pk_cast: typing.Callable | None = None,
        _stmt: sa.Select | None = None,
    ) -> None:
        self.pk_column = pk_column or guess_pk_field(model_class)
        self.pk_cast = pk_cast or guess_pk_type(model_class, self.pk_column)
        self.model_class = model_class
        self.query = query if query is not None else sa.select(model_class)
        self.query_for_list = query_for_list if query_for_list is not None else self.query
        self._stmt = _stmt if _stmt is not None else self.query

    @property
    def raw(self) -> sa.Select[T]:
        return self._stmt

    def get_query(self) -> SQLADataSource[T]:
        return self._clone(self.query)

    def get_query_for_index(self) -> SQLADataSource[T]:
        return self._clone(self.query_for_list)

    def get_pk(self, obj: T) -> str:
        return str(getattr(obj, self.pk_column))

    def new(self) -> T:
        return self.model_class()

    def apply_search(self, search_term: str, searchable_fields: typing.Sequence[str]) -> SQLADataSource[T]:
        if not search_term:
            return self._clone(self._stmt)

        clauses = []
        props = get_column_properties(self.model_class, searchable_fields)
        for prop in props.values():
            if len(prop.columns) > 1:  # pragma: no cover, no idea
                continue
            clauses.append(create_search_token(prop.columns[0], search_term))
        return self._clone(self._stmt.where(sa.or_(*clauses)))

    def apply_ordering(
        self,
        ordering: dict[str, SortingType],
        sortable_fields: typing.Sequence[str],
    ) -> SQLADataSource[T]:
        stmt = self._stmt.order_by(None)
        props = get_column_properties(self.model_class, sortable_fields)

        for ordering_field, ordering_dir in ordering.items():
            try:
                prop = props[ordering_field]
                if len(prop.columns) > 1:  # pragma: no cover
                    continue
                column = prop.columns[0]
                stmt = stmt.order_by(column.desc() if ordering_dir == 'desc' else column.asc())
            except KeyError:
                # ordering_field does not exist in properties, ignore
                continue
        return self._clone(stmt)

    def apply_string_filter(self, field: str, operation: StringOperation, value: str) -> SQLADataSource[T]:
        column = getattr(self.model_class, field)
        expr: sa.sql.ColumnElement = sa.func.lower(column)

        mapping = {
            StringOperation.exact: lambda stmt: stmt.where(expr == value),
            StringOperation.startswith: lambda stmt: stmt.where(expr.startswith(value)),
            StringOperation.endswith: lambda stmt: stmt.where(expr.endswith(value)),
            StringOperation.contains: lambda stmt: stmt.where(expr.like(f'%{value}%')),
            StringOperation.pattern: lambda stmt: stmt.where(expr.regexp_match(value)),
        }
        filter_ = mapping[operation]
        return self._clone(filter_(self._stmt))

    def apply_number_filter(
        self, field: str, operation: NumberOperation, value: int | float | decimal.Decimal
    ) -> SQLADataSource[T]:
        column = getattr(self.model_class, field)
        number_column = sa.sql.cast(column, sa.Integer)
        mapping = {
            NumberOperation.eq: lambda stmt: stmt.where(number_column == value),
            NumberOperation.gt: lambda stmt: stmt.where(number_column > value),
            NumberOperation.gte: lambda stmt: stmt.where(number_column >= value),
            NumberOperation.lt: lambda stmt: stmt.where(number_column < value),
            NumberOperation.lte: lambda stmt: stmt.where(number_column <= value),
        }
        filter_ = mapping[operation]
        return self._clone(filter_(self._stmt))

    def apply_date_filter(self, field: str, value: datetime.date) -> SQLADataSource[T]:
        column = getattr(self.model_class, field)
        return self._clone(self._stmt.where(column == value))

    def apply_date_range_filter(
        self, field: str, before: datetime.date | None, after: datetime.date | None
    ) -> SQLADataSource[T]:
        column = getattr(self.model_class, field)
        query = self
        if before:
            query = query._clone(query._stmt.where(column <= before))
        if after:
            query = query._clone(query._stmt.where(column >= after))
        return query

    def apply_choice_filter(self, field: str, choices: list[typing.Any], coerce: typing.Callable) -> SQLADataSource[T]:
        column = getattr(self.model_class, field)
        return self._clone(self._stmt.where(column.in_(choices)))

    def apply_boolean_filter(self, field: str, value: bool) -> SQLADataSource[T]:
        column = getattr(self.model_class, field)
        return self._clone(self._stmt.where(column.is_(value)))

    async def count(self, request: Request) -> int:
        stmt = sa.select(sa.func.count('*')).select_from(self._stmt)  # type: ignore[arg-type]
        result = await request.state.dbsession.scalars(stmt)
        return result.one()

    async def get(self, request: Request, pk: str) -> typing.Any:
        column = getattr(self.model_class, self.pk_column)
        stmt = self.get_query()._stmt.where(column == self.pk_cast(pk))
        result = await request.state.dbsession.scalars(stmt)
        return result.one()

    async def paginate(self, request: Request, page: int, page_size: int) -> Pagination[typing.Any]:
        row_count = await self.count(request)

        offset = (page - 1) * page_size
        stmt = self._stmt.limit(page_size).offset(offset)
        result = await request.state.dbsession.scalars(stmt)
        rows = result.all()
        return Pagination(rows=list(rows), total_rows=row_count, page=page, page_size=page_size)

    async def create(self, request: Request, model: typing.Any) -> None:
        request.state.dbsession.add(model)
        await request.state.dbsession.commit()

    async def update(self, request: Request, model: typing.Any) -> None:
        await request.state.dbsession.commit()

    async def delete(self, request: Request, *object_ids: str) -> None:
        typed_ids = list(map(self.pk_cast, object_ids))
        stmt = sa.delete(self.model_class).where(sa.column(self.pk_column).in_(typed_ids))
        async for instance in await request.state.dbsession.stream(stmt):
            await request.state.dbsession.delete(instance)
        await request.state.dbsession.commit()

    def _clone(self, stmt: sa.Select | None = None) -> SQLADataSource:
        return self.__class__(
            query=self.query,
            model_class=self.model_class,
            query_for_list=self.query_for_list,
            _stmt=stmt if stmt is not None else self._stmt,
        )

    def __repr__(self) -> str:  # pragma: no cover
        return repr(self._stmt)



from __future__ import annotations

import math
import typing
from starlette.requests import Request

M = typing.TypeVar('M')


def get_page_value(request: Request, param_name: str = 'page') -> int:
    page = 1
    try:
        page = max(1, int(request.query_params.get(param_name, 1)))
    except (TypeError, ValueError):
        pass
    return page


def get_page_size_value(request: Request, param_name: str = 'page_size', max_size: int = 100, default: int = 25) -> int:
    page_size = default
    try:
        page_size = int(request.query_params.get(param_name, default))
    except (TypeError, ValueError):
        pass
    return min(page_size, max_size)


class Pagination(typing.Generic[M]):
    def __init__(self, rows: typing.Sequence[M], total_rows: int, page: int, page_size: int) -> None:
        self.rows = rows
        self.total_rows = total_rows
        self.page = page
        self.page_size = page_size
        self._pointer = 0

    @property
    def total_pages(self) -> int:
        """Total pages in the row set."""
        return math.ceil(self.total_rows / self.page_size)

    @property
    def has_next(self) -> bool:
        """Test if the next page is available."""
        return self.page < self.total_pages

    @property
    def has_previous(self) -> bool:
        """Test if the previous page is available."""
        return self.page > 1

    @property
    def has_other(self) -> bool:
        """Test if page has next or previous pages."""
        return self.has_next or self.has_previous

    @property
    def next_page(self) -> int:
        """
        Next page number.

        Always returns an integer. If there is no more pages the current page number returned.
        """
        return min(self.total_pages, self.page + 1)

    @property
    def previous_page(self) -> int:
        """
        Previous page number.

        Always returns an integer. If there is no previous page, the number 1 returned.
        """
        return max(1, self.page - 1)

    @property
    def start_index(self) -> int:
        """The 1-based index of the first item on this page."""
        if self.page == 1:
            return 1
        return (self.page - 1) * self.page_size + 1

    @property
    def end_index(self) -> int:
        """The 1-based index of the last item on this page."""
        return min(self.start_index + self.page_size - 1, self.total_rows)

    def iter_pages(
        self, left_edge: int = 1, left_current: int = 1, right_current: int = 1, right_edge: int = 1
    ) -> typing.Generator[int | None, None, None]:
        pages_end = self.total_pages + 1

        if pages_end == 1:
            return

        left_end = min(1 + left_edge, pages_end)
        yield from range(1, left_end)

        if left_end == pages_end:
            return

        mid_start = max(left_end, self.page - left_current)
        mid_end = min(self.page + right_current + 1, pages_end)

        if mid_start - left_end > 0:
            yield None

        yield from range(mid_start, mid_end)

        if mid_end == pages_end:
            return

        right_start = max(mid_end, pages_end - right_edge)

        if right_start - mid_end > 0:
            yield None

        yield from range(right_start, pages_end)

    def __iter__(self) -> typing.Iterator[M]:
        return iter(self.rows)

    def __next__(self) -> M:
        if self._pointer == len(self.rows):
            raise StopIteration
        self._pointer += 1
        return self.rows[self._pointer - 1]

    def __getitem__(self, item: int) -> M:
        return self.rows[item]

    def __len__(self) -> int:
        return len(self.rows)

    def __bool__(self) -> bool:
        return len(self.rows) > 0

    def __str__(self) -> str:
        return (
            f'Page {self.page} of {self.total_pages}, rows {self.start_index} - {self.end_index} of {self.total_rows}.'
        )

    def __repr__(self) -> str:
        return f'<Page: page={self.page}, total_pages={self.total_pages}>'



from __future__ import annotations

import typing
from starlette.requests import HTTPConnection, Request
from starlette.responses import Response

if typing.TYPE_CHECKING:  # pragma:no cover
    from ohmyadmin.app import OhMyAdmin


def get_admin(conn: HTTPConnection) -> OhMyAdmin:
    return conn.state.admin


def render_to_string(
    request: Request,
    template_name: str,
    context: typing.Mapping[str, typing.Any] | None = None,
) -> str:
    return request.state.admin.render_to_string(request, template_name, context)


def render_to_response(
    request: Request,
    template_name: str,
    context: typing.Mapping[str, typing.Any] | None = None,
    headers: typing.Mapping[str, typing.Any] | None = None,
) -> Response:
    return request.state.admin.render_to_response(request, template_name, context, headers=headers)



import functools
import itertools
import jinja2
import operator
import os
import time
import typing
from async_storages import FileStorage, LocalStorage
from async_storages.file_server import FileServer
from markupsafe import Markup
from starlette import templating
from starlette.middleware import Middleware
from starlette.middleware.authentication import AuthenticationMiddleware
from starlette.requests import Request
from starlette.responses import RedirectResponse, Response
from starlette.routing import BaseRoute, Mount, Route, Router
from starlette.staticfiles import StaticFiles
from starlette.types import Receive, Scope, Send
from starlette_babel import gettext_lazy as _
from starlette_babel.contrib.jinja import configure_jinja_env
from starlette_flash import flash
from tabler_icons import tabler_icon

from ohmyadmin.authentication import AnonymousAuthPolicy, BaseAuthPolicy
from ohmyadmin.menu import MenuGroup, MenuLink, NavItem
from ohmyadmin.middleware import LoginRequiredMiddleware
from ohmyadmin.pages.base import BasePage
from ohmyadmin.templates import as_html_attrs, pk_filter

START_TIME = time.time()
PACKAGE_NAME = __name__.split('.')[0]


class OhMyAdmin(Router):
    def __init__(
        self,
        title: str = 'Oh My Admin!',
        logo_url: str = '',
        pages: typing.Sequence[BasePage] | None = None,
        template_dir: str | os.PathLike | None = None,
        file_storage: FileStorage | None = None,
        user_menu: typing.Sequence[NavItem] | None = None,
        auth_policy: BaseAuthPolicy = AnonymousAuthPolicy(),
    ) -> None:
        self.title = title
        self.pages = pages or []
        self.logo_url = logo_url
        self.auth_policy = auth_policy
        self.user_menu = user_menu or []
        self.file_storage = file_storage or FileStorage(LocalStorage('/tmp/ohmyadmin', mkdirs=True))
        self.middleware: list[Middleware] = [
            Middleware(AuthenticationMiddleware, backend=auth_policy.get_authentication_backend()),
            Middleware(LoginRequiredMiddleware, exclude_paths=['/login', '/static']),
        ]

        self.jinja_env = jinja2.Environment(
            autoescape=True,
            loader=jinja2.ChoiceLoader(
                [jinja2.FileSystemLoader(template_dir or '__undefined__'), jinja2.PackageLoader(PACKAGE_NAME)]
            ),
        )
        self.jinja_env.globals.update(
            {
                'admin': self,
                'tabler_icon': tabler_icon,
            }
        )
        self.jinja_env.filters.update(
            {
                'pk': pk_filter,
                'as_html_attrs': as_html_attrs,
            }
        )
        configure_jinja_env(self.jinja_env)

        self.templates = templating.Jinja2Templates('')
        self.templates.env = self.jinja_env

        super().__init__(self.get_routes())

    def url_for(self, request: Request, path_name: str, **path_params: typing.Any) -> str:
        """
        Generate URL to a named route.

        This function generates a correct URL when used in a mounted application (e. g. Mount()).
        """
        return request.url_for(path_name, **path_params)

    def static_url(self, request: Request, path: str) -> str:
        """Generate URL to a static asset managed by the admin application."""
        _suffix = time.time() if request.app.debug else START_TIME
        return request.url_for('ohmyadmin.static', path=path) + f'?{_suffix}'

    def media_url(self, request: Request, path: str) -> str:
        """
        Generate URL to a media (uploaded) file.

        If `path` starts with `http` or `https` then the value used as is. Otherwise, generates URL to a media server
        which internally calls `file_storage` to generate the URL. The actual URL generation logic delegated to
        `file_storage`.
        """
        if path.startswith('http://') or path.startswith('https://'):
            return path
        return request.url_for('ohmyadmin.media', path=path)

    def render_to_string(
        self,
        request: Request,
        template_name: str,
        context: typing.Mapping[str, typing.Any] | None = None,
    ) -> str:
        context = dict(context or {})
        context.update(
            {
                'request': request,
                'url': functools.partial(self.url_for, request),
                'static_url': functools.partial(self.static_url, request),
                'media_url': functools.partial(self.media_url, request),
                'flash_messages': flash(request),
            }
        )
        return Markup(self.templates.get_template(template_name).render(context))

    def render_to_response(
        self,
        request: Request,
        template_name: str,
        context: typing.Mapping[str, typing.Any] | None = None,
        headers: typing.Mapping[str, typing.Any] | None = None,
    ) -> Response:
        context = dict(context or {})
        context.update(
            {
                'request': request,
                'url': functools.partial(self.url_for, request),
                'static_url': functools.partial(self.static_url, request),
                'media_url': functools.partial(self.media_url, request),
                'flash_messages': flash(request),
            }
        )
        return self.templates.TemplateResponse(template_name, context, headers=headers)

    def get_routes(self) -> typing.Sequence[BaseRoute]:
        return [
            Route('/', self.index_view, name='ohmyadmin.welcome'),
            Route('/login', self.login_view, name='ohmyadmin.login', methods=['GET', 'POST']),
            Route('/logout', self.logout_view, name='ohmyadmin.logout', methods=['POST']),
            Mount('/static', StaticFiles(packages=['ohmyadmin']), name='ohmyadmin.static'),
            Mount('/media', FileServer(self.file_storage), name='ohmyadmin.media'),
            Mount('/', routes=[page.as_route() for page in self.pages]),
        ]

    def index_view(self, request: Request) -> Response:
        return self.render_to_response(
            request, 'ohmyadmin/index.html', {'page_title': _('Welcome', domain='ohmyadmin')}
        )

    async def login_view(self, request: Request) -> Response:
        next_url = request.query_params.get('next', request.url_for('ohmyadmin.welcome'))
        form_class = self.auth_policy.get_login_form_class()
        form = form_class(formdata=await request.form(), data={'next_url': next_url})
        if request.method in ['POST'] and form.validate():
            if user := await self.auth_policy.authenticate(request, form.identity.data, form.password.data):
                self.auth_policy.login(request, user)
                flash(request).success(_('You have been logged in.', domain='ohmyadmin'))
                return RedirectResponse(url=form.next_url.data, status_code=302)
            else:
                flash(request).error(_('Invalid credentials.', domain='ohmyadmin'))

        return self.render_to_response(
            request,
            'ohmyadmin/login.html',
            {
                'request': request,
                'form': form,
                'next_url': next_url,
                'page_title': _('Login', domain='ohmyadmin'),
            },
        )

    async def logout_view(self, request: Request) -> Response:
        self.auth_policy.logout(request)
        flash(request).success(_('You have been logged out.', domain='ohmyadmin'))
        return RedirectResponse(request.url_for('ohmyadmin.login'), status_code=302)

    def get_user_menu(self, request: Request) -> typing.Sequence[NavItem]:
        user_menu = self.user_menu
        return user_menu

    def get_main_menu(self, request: Request) -> list[NavItem]:
        groups = itertools.groupby(
            self.pages,
            key=operator.attrgetter('group'),
        )
        return [
            MenuGroup(
                text=group[0],
                items=[
                    MenuLink(
                        icon=item.icon,
                        text=item.label_plural,
                        url=item.generate_url(request),
                    )
                    for item in group[1]
                ],
            )
            for group in groups
        ]

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        scope.setdefault('state', {})
        scope['state'].update({'admin': self})

        app = super().__call__
        for middleware in reversed(self.middleware):
            app = middleware.cls(app, **middleware.options)

        await app(scope, receive, send)



import datetime
import decimal
import functools
import re
import typing
from starlette.requests import Request

from ohmyadmin.datasource.base import DataSource, NumberOperation, StringOperation
from ohmyadmin.ordering import SortingType
from ohmyadmin.pagination import Pagination

T = typing.TypeVar('T')


def apply_search_filter(current_obj: typing.Any, field: str, term: str) -> bool:
    field_value = getattr(current_obj, field, term).lower()
    search_term = term.lower()[1:]

    if term.startswith('^'):
        return field_value.startswith(search_term)
    if term.startswith('$'):
        return field_value.endswith(search_term)
    if term.startswith('='):
        return field_value == search_term
    if term.startswith('@'):
        return re.search(search_term, field_value) is not None
    return search_term in field_value


def search_filter(obj: typing.Any, fields: typing.Sequence[str], term: str) -> bool:
    return all([apply_search_filter(obj, field, term) for field in fields])


class InMemoryDataSource(DataSource[T]):
    def __init__(self, model_class: type[T], objects: typing.Sequence[T]) -> None:
        self.objects = list(objects)
        self.model_class = model_class
        self._generate_cache()

    def get_query_for_index(self) -> DataSource:
        return self

    def get_pk(self, obj: typing.Any) -> str:
        return str(obj.id)

    def clone(self, objects: typing.Sequence[T] | None = None) -> DataSource[T]:
        return InMemoryDataSource(
            model_class=self.model_class,
            objects=self.objects if objects is None else objects,
        )

    def apply_search(self, search_term: str, searchable_fields: typing.Sequence[str]) -> DataSource[T]:
        filtered = [obj for obj in self.objects if search_filter(obj, searchable_fields, search_term)]
        return self.clone(objects=filtered)

    def apply_ordering(self, ordering: dict[str, SortingType], sortable_fields: typing.Sequence[str]) -> DataSource[T]:
        def get_key(obj: typing.Any, field: str) -> str:
            return getattr(obj, field)

        filtered = self.objects.copy()
        for field, direction in ordering.items():
            if field not in sortable_fields:
                continue
            filtered = sorted(self.objects, key=functools.partial(get_key, field=field), reverse=direction == 'desc')
        return self.clone(objects=filtered)

    def apply_string_filter(self, field: str, operation: StringOperation, value: str) -> DataSource[T]:
        value = value.lower()

        def get_value(obj: typing.Any, attr: str) -> str:
            return getattr(obj, attr, '').lower()

        assert operation in StringOperation
        match operation:
            case StringOperation.startswith:
                return self.clone(objects=[obj for obj in self.objects if get_value(obj, field).startswith(value)])
            case StringOperation.endswith:
                return self.clone(objects=[obj for obj in self.objects if get_value(obj, field).endswith(value)])
            case StringOperation.contains:
                return self.clone(objects=[obj for obj in self.objects if value in get_value(obj, field)])
            case StringOperation.pattern:
                return self.clone(objects=[obj for obj in self.objects if re.match(value, get_value(obj, field))])
            case StringOperation.exact:
                return self.clone(objects=[obj for obj in self.objects if get_value(obj, field) == value])

    def apply_number_filter(
        self, field: str, operation: NumberOperation, value: int | float | decimal.Decimal
    ) -> DataSource[T]:
        def get_value(obj: typing.Any, attr: str) -> int | float | decimal.Decimal:
            return getattr(obj, attr, 0)

        assert operation in NumberOperation
        match operation:
            case NumberOperation.eq:
                return self.clone(objects=[obj for obj in self.objects if get_value(obj, field) == value])
            case NumberOperation.gt:
                return self.clone(objects=[obj for obj in self.objects if get_value(obj, field) > value])
            case NumberOperation.gte:
                return self.clone(objects=[obj for obj in self.objects if get_value(obj, field) >= value])
            case NumberOperation.lt:
                return self.clone(objects=[obj for obj in self.objects if get_value(obj, field) < value])
            case NumberOperation.lte:
                return self.clone(objects=[obj for obj in self.objects if get_value(obj, field) <= value])

    def apply_date_filter(self, field: str, value: datetime.date | datetime.datetime) -> DataSource[T]:
        return self.clone(objects=[obj for obj in self.objects if getattr(obj, field, None) == value])

    def apply_date_range_filter(
        self,
        field: str,
        before: datetime.date | datetime.datetime | None,
        after: datetime.date | datetime.datetime | None,
    ) -> DataSource[T]:
        def check_before(obj: typing.Any) -> bool:
            value = getattr(obj, field, None)
            if value is None:
                return False
            return value <= before

        def check_after(obj: typing.Any) -> bool:
            value = getattr(obj, field, None)
            if value is None:
                return False
            return value >= after

        return self.clone(
            objects=[
                obj
                for obj in self.objects
                if all(
                    [
                        check_before(obj) if before else True,
                        check_after(obj) if after else True,
                    ]
                )
            ]
        )

    def apply_choice_filter(self, field: str, choices: list[typing.Any], coerce: typing.Callable) -> DataSource[T]:
        def get_value(obj: typing.Any, attr: str) -> typing.Any:
            return coerce(getattr(obj, attr, ''))

        return self.clone(objects=[obj for obj in self.objects if get_value(obj, field) in choices])

    def apply_boolean_filter(self, field: str, value: bool) -> DataSource[T]:
        return self.clone(objects=[obj for obj in self.objects if getattr(obj, field) is value])

    async def get(self, request: Request, pk: str) -> T | None:
        return self._by_pk.get(pk)

    async def paginate(self, request: Request, page: int, page_size: int) -> Pagination[T]:
        start_offset = max(0, page - 1) * page_size
        end_offset = start_offset + page_size
        result = self.objects[start_offset:end_offset]
        return Pagination(rows=result, total_rows=len(self.objects), page=page, page_size=page_size)

    async def create(self, request: Request, model: typing.Any) -> None:
        self.objects.append(model)
        self._generate_cache()

    async def delete(self, request: Request, *object_ids: str) -> None:
        for object_id in object_ids:
            self.objects.remove(self._by_pk[object_id])
        self._generate_cache()

    def new(self) -> typing.Any:
        return self.model_class()

    async def update(self, request: Request, model: typing.Any) -> None:
        pass

    def _generate_cache(self) -> None:
        self._by_pk: dict[str, T] = {self.get_pk(obj): obj for obj in self.objects}



from __future__ import annotations

import abc
import typing
from starlette.datastructures import URL
from starlette.requests import Request

from ohmyadmin.helpers import LazyURL, resolve_url
from ohmyadmin.shortcuts import render_to_string


class NavItem(abc.ABC):  # pragma: no cover
    @abc.abstractmethod
    def is_active(self, request: Request) -> bool:
        ...

    @abc.abstractmethod
    def render(self, request: Request) -> str:
        ...


class MenuLink(NavItem):
    """
    Menu item for use in primary navigation (e.

    g. in a sidebar).
    """

    def __init__(self, text: str, url: str | LazyURL, icon: str = '') -> None:
        self.text = text
        self.icon = icon
        self.url = url

    def is_active(self, request: Request) -> bool:
        return request.url.path.startswith(resolve_url(request, self.url).path)

    def resolve(self, request: Request) -> URL:
        return resolve_url(request, self.url)

    def render(self, request: Request) -> str:
        return render_to_string(
            request,
            'ohmyadmin/menu_item_link.html',
            {
                'item': self,
                'is_active': self.is_active(request),
            },
        )


class MenuGroup(NavItem):
    def __init__(self, text: str, items: list[MenuLink], icon: str = '') -> None:
        self.text = text
        self.icon = icon
        self.items = items

    def __iter__(self) -> typing.Iterator[MenuLink]:
        yield from self.items

    def render(self, request: Request) -> str:
        return render_to_string(request, 'ohmyadmin/menu_group.html', {'group': self})

    def is_active(self, request: Request) -> bool:
        return any([item.is_active(request) for item in self.items])
